# channels seem be fixed --> cool
# finalize whsed and corrections --> make sure they are channel dependent
# check which channels are being restored by the new wshed within the paint arena
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
# in set image add sliders for D and Z --> see how I can do that --> or do it directly in paint ????
import traceback

import matplotlib.pyplot as plt
import numpy as np
from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import QSize, Qt
from qtpy.QtGui import QPalette, QKeySequence, QPainter
from qtpy.QtWidgets import QScrollArea, QVBoxLayout, QWidget, QSpinBox, QComboBox, QToolBar, QStatusBar, QLabel, \
    QHBoxLayout, QAction, QSlider
import qtawesome as qta
from qtpy.QtCore import Qt, QTimer
from epyseg.draw.shapes.freehand2d import Freehand2D
from epyseg.draw.shapes.image2d import Image2D
from epyseg.draw.shapes.rect2d import Rect2D
from epyseg.img import Img, int24_to_RGB, has_metadata
from epyseg.ta.GUI.paint2 import Createpaintwidget

# TODO --> could maybe add icons just below scroll if needed to add actions
# j'arrive pas à creer un truc dejà self embedded
# make it also handle other stuff

# --> nb replace the Createpaintwidget in multi_image_display.py by this --> better in fact
# shall I do full screen ???
# TODO handle channels too if needed
# could store mask and memory of drawing here too --> maybe on mouse release ???
# avec des ctrl+Z et des ctrl+R


# need zooms
#

class scrollable_paint(QWidget):

    def __init__(self, custom_paint_panel=None):
        super().__init__()
        if custom_paint_panel is None:
            self.paint = Createpaintwidget()
        else:
            self.paint = custom_paint_panel
        layout = QVBoxLayout()

        # delay dimensionality change update --> just to make sure it does not overwhelms the system upon slider change
        # think about how to do that in a smart way -> maybe by changing the connect ???
        # self.delayed_dimension_preview = QTimer()
        # self.delayed_dimension_preview.setSingleShot(True)
        # self.delayed_dimension_preview.timeout.connect(self.slider_dimension_value_changed)
        self.dimension_sliders = []
        self.is_width_for_alternating = True
        # add the possibility to browse dimensions
        self.dimensions_container = QVBoxLayout() # this will contain all the browsable dimensions of the image
        # fake add sliders here
        # self.dim_slider_with_label1 = self.create_dim_slider(dimension='d', max_dim=16)
        # self.dimensions_container.addLayout(self.dim_slider_with_label1)
        # self.dim_slider_with_label2 = self.create_dim_slider(dimension='c', max_dim=3)
        # self.dimensions_container.addLayout(self.dim_slider_with_label2)
        # self.dim_slider_with_label3 = self.create_dim_slider(dimension='t', max_dim=22)
        # self.dimensions_container.addLayout(self.dim_slider_with_label3)
        # each of the sliders will be assoc to a single dim --> TODO
        # maybe remove the slider for channels ???
        # same as the other with embed in scroll area
        self.scrollArea = QScrollArea()
        self.scrollArea.setBackgroundRole(QPalette.Dark)
        self.scrollArea.setWidget(self.paint)
        self.paint.scrollArea = self.scrollArea
        layout.addLayout(self.dimensions_container)
        layout.addWidget(self.scrollArea)
        self.setLayout(layout)
        status_bar = QStatusBar()
        layout.addWidget(status_bar)
        self.paint.statusBar = status_bar
        self.add_shortcuts()
        self.drawing_commands()

    # this guy should handle the shortcuts it can and pass the rest to the other layer
    def add_shortcuts(self):
        # print('SHORTCUTS ADDED')

        # if True:
        #     return
        zoomPlus = QtWidgets.QShortcut("Ctrl+Shift+=", self)
        zoomPlus.activated.connect(self.zoomIn)
        zoomPlus.setContext(QtCore.Qt.ApplicationShortcut)



        # that works and disconnects all
        # zoomPlus.disconnect()



        zoomPlus2 = QtWidgets.QShortcut("Ctrl++", self)
        zoomPlus2.activated.connect(self.zoomIn)
        zoomPlus2.setContext(QtCore.Qt.ApplicationShortcut)

        zoomMinus = QtWidgets.QShortcut("Ctrl+Shift+-", self)
        zoomMinus.activated.connect(self.zoomOut)
        zoomMinus.setContext(QtCore.Qt.ApplicationShortcut)

        zoomMinus2 = QtWidgets.QShortcut("Ctrl+-", self)
        zoomMinus2.activated.connect(self.zoomOut)
        zoomMinus2.setContext(QtCore.Qt.ApplicationShortcut)

        ctrl0 = QtWidgets.QShortcut("Ctrl+0", self)
        ctrl0.activated.connect(self.zoom_reset)
        ctrl0.setContext(QtCore.Qt.ApplicationShortcut)

        self.ctrlS = QtWidgets.QShortcut("Ctrl+S", self)
        self.ctrlS.activated.connect(self.paint.save)
        self.ctrlS.setContext(QtCore.Qt.ApplicationShortcut)


        self.ctrlM = QtWidgets.QShortcut("Ctrl+M", self)
        self.ctrlM.activated.connect(self.paint.ctrl_m_apply)
        self.ctrlM.setContext(QtCore.Qt.ApplicationShortcut)

        self.shrtM = QtWidgets.QShortcut("M", self)
        self.shrtM.activated.connect(self.paint.m_apply)
        self.shrtM.setContext(QtCore.Qt.ApplicationShortcut)

        self.increase_contrastC = QtWidgets.QShortcut('C', self)
        self.increase_contrastC.activated.connect(self.paint.increase_contrast)
        self.increase_contrastC.setContext(QtCore.Qt.ApplicationShortcut)

        # I can connect to progeny --> makes sense and really easy --> cool
        self.enterShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Return), self)
        self.enterShortcut.activated.connect(self.paint.apply)
        self.enterShortcut.setContext(QtCore.Qt.ApplicationShortcut)

        self.enterShortcut2 = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Enter), self)
        self.enterShortcut2.activated.connect(self.paint.apply)
        self.enterShortcut2.setContext(QtCore.Qt.ApplicationShortcut)

        # en fait je peux aussi connecter des trucs au descendant --> pas bete et tres facile à faire!!!

        # pkoi ça ne marche pas !!!
        self.shiftEnterShortcut = QtWidgets.QShortcut("Shift+Enter", self)
        self.shiftEnterShortcut.activated.connect(self.paint.shift_apply)
        self.shiftEnterShortcut.setContext(QtCore.Qt.ApplicationShortcut)
        #
        self.shiftEnterShortcut2 = QtWidgets.QShortcut("Shift+Return", self)
        self.shiftEnterShortcut2.activated.connect(self.paint.shift_apply)
        self.shiftEnterShortcut2.setContext(QtCore.Qt.ApplicationShortcut)

        self.ctrl_shift_S_grab_screen_shot = QtWidgets.QShortcut('Ctrl+Shift+S', self)
        self.ctrl_shift_S_grab_screen_shot.activated.connect(self.paint.grab_screen_shot)
        self.ctrl_shift_S_grab_screen_shot.setContext(QtCore.Qt.ApplicationShortcut)

        self.supr = QtWidgets.QShortcut(QKeySequence(Qt.Key_Delete), self)
        self.supr.activated.connect(self.paint.suppr_pressed)
        self.supr.setContext(QtCore.Qt.ApplicationShortcut)



        # peut etre faire un fullscreen shortcut --> can be useful

    def enable_shortcuts(self):
        self.disable_shortcuts()
        self.ctrlS.activated.connect(self.paint.save)
        self.ctrlM.activated.connect(self.paint.ctrl_m_apply)
        self.shrtM.activated.connect(self.paint.m_apply)
        self.enterShortcut.activated.connect(self.paint.apply)
        self.enterShortcut2.activated.connect(self.paint.apply)
        self.shiftEnterShortcut.activated.connect(self.paint.shift_apply)
        self.shiftEnterShortcut2.activated.connect(self.paint.shift_apply)

    def disable_shortcuts(self):
        # TODO
        try:
            self.ctrlS.disconnect()
        except:
            pass
        # TODO
        try:
            self.ctrlM.disconnect()
        except:
            pass
        try:
            self.shrtM.disconnect()
        except:
            pass
        try:
            self.enterShortcut.disconnect()
        except:
            pass
        try:
            self.enterShortcut2.disconnect()
        except:
            pass
        try:
            self.shiftEnterShortcut.disconnect()
        except:
            pass
        try:
            self.shiftEnterShortcut2.disconnect()
        except:
            pass

    def enableMouseTracking(self):
        self.paint.drawing_enabled = True

    def disableMouseTracking(self):
        self.paint.drawing_enabled = False


    # def gather_indices_for_all_dimensions(self):
        # loop through the content of an image ????



    # ça marche --> ajouter fit to width ou fit to height
    def fit_to_width_or_height(self):
        if self.is_width_for_alternating:
            self.fit_to_width()
        else:
            self.fit_to_height()
        self.is_width_for_alternating = not self.is_width_for_alternating

    # almost there but small bugs
    def fit_to_width(self):
        if self.paint.image is None:
            return
        # self.scrollArea = self.scroll_areas[nb]
        width = self.scrollArea.width() - 2
        width -= self.scrollArea.verticalScrollBar().sizeHint().width()
        # height-=self.scrollArea.horizontalScrollBar().sizeHint().height()
        width_im = self.paint.image.width()
        scale = width / width_im
        self.scaleImage(scale)

    def fit_to_height(self, nb=0):
        # paint = self.paints[nb]
        if self.paint.image is None:
            return
        # scrollArea = self.scroll_areas[nb]
        height = self.scrollArea.height() - 2
        height -= self.scrollArea.horizontalScrollBar().sizeHint().height()
        height_im = self.paint.image.height()
        scale = height / height_im
        self.scaleImage(scale)

    def fit_to_window(self):
        # compute best fit
        # just get the size of the stuff and best fit it
        # paint = self.paints[nb]
        if self.paint.image is None:
            return

        # QScrollArea.ensureWidgetVisible(
        # QScrollArea.ensureVisible
        # QScrollArea.setWidgetResizable
        # scrollArea = self.scroll_areas[nb]
        width = self.scrollArea.width() - 2  # required to make sure bars not visible
        height =self. scrollArea.height() - 2

        # scale image that it fits in --> change scale

        # width-=self.scrollArea.verticalScrollBar().sizeHint().width()
        # height-=self.scrollArea.horizontalScrollBar().sizeHint().height()

        height_im = self.paint.image.height()
        width_im = self.paint.image.width()
        scale = height / height_im
        if width / width_im < scale:
            scale = width / width_im
        self.scaleImage(scale)




    def zoomIn(self):
        # 10% more each time
        self.scaleImage(self.paint.scale + (self.paint.scale*10.0/100.0))# une hausse de 10% à chaque fois

    def zoomOut(self):
        # 10% less each time
        self.scaleImage(self.paint.scale - (self.paint.scale*10.0/100.0))

    def zoom_reset(self):
        # resets zoom
        self.scaleImage(1.0)

    # pb this is redundant with the main GUI and not in sync --> can cause a lot of trouble --> dactivate for now
    def full_screen(self):
            if not self.isFullScreen():
                self.fullscreen.setIcon(qta.icon('mdi.fullscreen-exit'))
                self.fullscreen.setToolTip('Exit full screen')
                self.showFullScreen()
                # self.setWindowFlags(
                #     QtCore.Qt.Window |
                #     QtCore.Qt.CustomizeWindowHint |
                #     # QtCore.Qt.WindowTitleHint |
                #     # QtCore.Qt.WindowCloseButtonHint |
                #     QtCore.Qt.WindowStaysOnTopHint
                # )
            else:
                # settings = QSettings()
                # self.Stack.restoreGeometry(settings.value("geometry")) #.toByteArray()
                self.fullscreen.setIcon(qta.icon('mdi.fullscreen'))
                self.fullscreen.setToolTip('Enter full screen')
                self.setWindowFlags(QtCore.Qt.Widget)
                # # self.Stack.setWindowFlags(self.flags)
                # self.grid.addWidget(self.Stack, 0,
                #                     0)  # pas trop mal mais j'arrive pas à le remettre dans le truc principal
                # # dirty hack to make it repaint properly --> obviously not all lines below are required but some are --> need test, the last line is key though
                # self.grid.update()
                # self.Stack.update()
                # self.Stack.show()
                # self.centralWidget().setLayout(self.grid)
                # self.centralWidget().update()
                # self.update()
                self.show()
                self.repaint()
                # self.Stack.update()
                # self.Stack.repaint()
                # self.centralWidget().repaint()

    def scaleImage(self, scale):
        self.paint.set_scale(scale)
        if self.paint.image is not None:
            self.paint.resize(self.paint.scale * self.paint.image.size())
        else:
            self.paint.resize(QSize(0, 0))
        self.paint.update()

    # can maybe add pen stuff here --> simpler and easier
    # need zoom buttons here too
    '''
    buttons/icons to add at some point
    
        
        "Cell smaller than this value will be destroyed"
        "small_cell_remover"
        
        
        "/Icons/refresh.png"
        "run 2 seeds watershed (click twice in a poorly segmented cell to have a the bond be created using the watershed algorithm)"
        
        "/Icons/Knob Snapback.png"
        "Undo"
        
        "/Icons/Knob Valid Blue.png"
        "remove small cells and apply correction"
        
        "/Icons/Knob Valid Green.png"
        "apply correction"
        
        "/Icons/channels.png"
        "Channel Selection"
        "Channel to use for Watershed, polarity, recentering,...")
        
        "/Icons/onebit_11.png"
        "Save"
        
        
        "/Icons/eye.png"
        "View below the mask (lasts 2s)"
        
        
        "/Icons/zoom in.png"
        "Zoom +"
        
        "/Icons/zoom out.png"
        "Zoom -"
        
        "/Icons/1in1.png"
        "1:1 aspect ratio"
        
        "/Icons/fit_2_screen.gif"
        "Fit to panel"
        
        
        "/Icons/double_diagonal_arrow.gif"
        "Alternate between best fit in height or in width"
        
        
        "/Icons/Import Picture Document.png"
        "Send Current View To The System Clipboard"
        
        "/Icons/pipet_small.png"
        "Color pipet"
        
        "Mask col"
        "mask color"
        
        "Correct Drift" 
    
    '''

    # add shortcut for show hide mask and maybe also full screen --> not a good idea

    def drawing_commands(self):
        self.penSize = QSpinBox(objectName='penSize')
        self.penSize.setToolTip('Change this value to change pen size')
        self.penSize.setSingleStep(1)
        self.penSize.setRange(1, 256)
        self.penSize.setValue(3)
        self.penSize.valueChanged.connect(self.penSizechange)

        self.channels = QComboBox(objectName='channels')
        self.channels.setToolTip('Select the channel of interest')
        self.channels.addItem("merge")
        # self.channels.addItem("0")
        # self.channels.addItems(["1", "2", "3"])
        self.channels.currentIndexChanged.connect(self.channelChange)


        hlayout_of_toolbars = QHBoxLayout()
        hlayout_of_toolbars.setAlignment(Qt.AlignLeft)
        self.tb1 = QToolBar()
        self.tb2 = QToolBar()
        self.tb3 = QToolBar()


        pencil_action = QAction(qta.icon('ei.pencil'), 'Pen Size', self)
        pencil_action.setEnabled(False)
        self.tb2.addAction(pencil_action)
        # tb.addAction("sq...")
        self.tb2.addWidget(self.penSize)

        # toolButton = QToolButton()
        # toolButton.setText("Draw")
        # toolButton.setIcon(self.style().standardIcon(QStyle.SP_DriveFDIcon))
        # tb.addWidget(toolButton)

        # fa5.save
        save_action = QAction(qta.icon('fa5.save'), 'Save mask', self)
        save_action.triggered.connect(self.paint.save)
        self.tb2.addAction(save_action)


        # toolButton2 = QToolButton()
        # toolButton2.setText("Draw")
        # toolButton2.setIcon(self.style().standardIcon(QStyle.SP_DialogApplyButton))
        # tb.addWidget(toolButton2)


        show_hide_mask =QAction(qta.icon('fa.bullseye'), 'Show/hide mask (same as pressing "M")', self)
        show_hide_mask.triggered.connect(self.paint.m_apply)
        self.tb2.addAction(show_hide_mask)

        apply_drawing = QAction(qta.icon('fa.check'), 'Apply drawing (same as pressing the "Enter" key)', self)
        apply_drawing.triggered.connect(self.paint.apply)
        self.tb2.addAction(apply_drawing)

        # toolButton3 = QToolButton()
        # toolButton3.setText("Draw")
        # toolButton3.setIcon(self.style().standardIcon(QStyle.SP_DialogCancelButton))
        # tb.addWidget(toolButton3)

        # toolButton4 = QToolButton()
        # toolButton4.setText("Draw")
        # toolButton4.setIcon(self.style().standardIcon(QStyle.SP_DialogCloseButton))
        # tb.addWidget(toolButton4)

        # toolButton5 = QToolButton()
        # toolButton5.setText("Draw")
        # toolButton5.setIcon(self.style().standardIcon(QStyle.SP_TrashIcon))
        # tb.addWidget(toolButton5)

        # toolButton6 = QToolButton()
        # toolButton6.setText("Draw")
        # toolButton6.setIcon(self.style().standardIcon(QStyle.SP_ArrowBack))
        # tb.addWidget(toolButton6)

        # toolButton7 = QToolButton()
        # toolButton7.setText("Draw")
        # toolButton7.setIcon(self.style().standardIcon(QStyle.SP_BrowserReload))
        # tb.addWidget(toolButton7)

        self.small_cell_size = QSpinBox()
        self.small_cell_size.setToolTip('Select the minimal pixel area for a cell to be considered as a cell.\nAny cell having an area less than this value will be removed.')
        self.small_cell_size.setSingleStep(1)
        self.small_cell_size.setRange(1, 1000000)
        self.small_cell_size.setValue(10)
        self.small_cell_size.valueChanged.connect(self.small_cell_size_changed)

        self.tb2.addWidget(self.small_cell_size)
        apply_rm_small_cells = QAction(qta.icon('mdi.check-underline'), 'Apply drawing and remove small cells (same as pressing "Shift+Enter")', self)
        apply_rm_small_cells.triggered.connect(self.paint.shift_apply)# self.paint.shift_apply
        self.tb2.addAction(apply_rm_small_cells)





        # marche pas --> comment faire pr que ça marche...
        local_reshed = QAction(qta.icon('mdi.reload'), 'Locally seeded watershed (same as pressing "Ctrl/Cmd + M")', self)
        local_reshed.triggered.connect(self.paint.ctrl_m_apply)# self.paint.shift_apply
        self.tb2.addAction(local_reshed)
        # hgjhghj

        # toolButton8 = QToolButton()
        # toolButton8.setText("Zoom+")
        # toolButton8.setIcon(qta.icon('ei.zoom-in')) # mdi.magnify-plus-outline --> marche aussi
        # tb.addWidget(toolButton8)

        # toolButton9 = QToolButton()
        # toolButton9.setText("Zoom-")
        # toolButton9.setIcon(qta.icon('ei.zoom-out'))
        # tb.addWidget(toolButton9)

        zoom_plus = QAction(qta.icon('ei.zoom-in'), 'Zoom+', self)
        zoom_plus.triggered.connect(self.zoomIn)
        self.tb1.addAction(zoom_plus)

        zoom_minus = QAction(qta.icon('ei.zoom-out'), 'Zoom-', self)
        zoom_minus.triggered.connect(self.zoomOut)
        self.tb1.addAction(zoom_minus)


        zoom_width_or_height = QAction(qta.icon('mdi.fit-to-page-outline'), 'Alternates between best fit in width and height', self)
        zoom_width_or_height.triggered.connect(self.fit_to_width_or_height)
        self.tb1.addAction(zoom_width_or_height)

        zoom_0 = QAction(qta.icon('ei.resize-full'), 'Reset zoom', self)
        zoom_0.triggered.connect(self.zoom_reset)
        self.tb1.addAction(zoom_0)

        # redundant with TA full screen and not in sync --> disable for now
        # self.fullscreen = QtWidgets.QAction(qta.icon('mdi.fullscreen'), 'Full screen', self)
        # self.fullscreen.setToolTip('Enter full screen')
        # self.fullscreen.triggered.connect(self.full_screen)
        # self.tb1.addAction(self.fullscreen)




        # exit_fullscreen = QtWidgets.QAction(qta.icon('mdi.fullscreen'), 'Full screen', self)
        # exit_fullscreen.triggered.connect(self.showFullScreen)
        # self.tb1.addAction(fullscreen)



        self.channel_label = QLabel('Channels:')
        self.channel_label.setToolTip('Select the channel of interest')
        self.tb3.addWidget(self.channel_label)
        self.tb3.addWidget(self.channels)

        # mdi.fullscreen
        # mdi.fullscreen-exit
        # ei.fullscreen --> a bit less good I find
        # fa.bullseye

        # could have show and hide mask here --> see how I can do that --> but do not delete it in fact

        # self.layout().addWidget(self.tb1)
        hlayout_of_toolbars.addWidget(self.tb2)


        # self.tb2.addAction('toto')
        # self.tb2.addAction('ttutu')
        # self.tb2.addAction('tata')
        hlayout_of_toolbars.addWidget(self.tb3)


        # one for the zooms
        # one for drawing and apply different things
        # TODO do several toolbars to deactivate different levels of stuff
        # and only allow save in the appropriate stuff

        # self.tb3.addAction('toto')
        # self.tb3.addAction('ttutu')
        # self.tb3.addAction('tata')
        hlayout_of_toolbars.addWidget(self.tb1)

        self.layout().addLayout(hlayout_of_toolbars)

    def get_selected_channel(self):
        idx = self.channels.currentIndex()
        if idx == 0:
            return None
        else:
            return idx-1

    # TODO also implement channel change directly within the display tool
    # if merge is applied --> apply on average of all channels --> maybe not so smart an idea but ok to start with and better than what I do in TA
    def channelChange(self, i):
        # update displayed image depending on channel
        # dqqsdqsdqsd
        # pass
        # try change channel if


        # tODO --> need at least to reactivate a bit that
        # pass
        # print('in channel change') # needs a fix
        meta = None
        try:
            meta = self.paint.raw_image.metadata
        except:
            pass
        self.paint.set_display(self.get_image_to_display_including_all_dims(), metadata=meta)
        self.paint.channelChange(i, skip_update_display=True)

        # print('in channel change !!!')
        # if self.img is not None:
        #     # print('in', self.img.metadata)
        #     if self.Stack.currentIndex() == 0:
        #         # need copy the image --> implement that
        #         # print(self.img[..., i].copy())
        #         # print(self.img[..., i])
        #         if i == 0:
        #             self.paint.setImage(self.img)
        #             # print('original', self.img.metadata)
        #         else:
        #             # print('modified0', self.img.metadata)
        #             channel_img = self.img.imCopy(c=i - 1)  # it's here that it is affected
        #             # print('modified1', self.img.metadata)
        #             # print('modified2', channel_img.metadata)
        #             self.paint.setImage(channel_img)
        #         self.paint.update()
        # else:
        #     # logger.error("Not implemented yet TODO add support for channels in 3D viewer")
        #     # sdqdqsdsqdqsd
        #     self.loadVolume()
        # or reimplement that for multi channels !!!



    # TODO maybe handle pen size change directly within the display --> can be useful by the way
    def small_cell_size_changed(self):
        self.paint.minimal_cell_size = self.small_cell_size.value()

    # TODO maybe handle pen size change directly within the display --> can be useful by the way
    def penSizechange(self):
        self.paint.brushSize = self.penSize.value()
        # self.update() # too slow if I update --> avoid
        # self.l1.setFont(QFont("Arial", size))

    def set_image(self, img):

        # probably I need store raw image to avoid issues --> TODO ???
        self.paint.set_image(img)# bug seems to be here

        img = self.paint.raw_image

        self.update_image_dimensions()

        # need also update the dimensions if any

        # print(type(img))

        # else I need also reset channels

        # if img is not None:
        #     if img.has_c():
        self._update_channels(img=img)
        # print('out')
        # channels = self.paint.get_nb_channels()
        # self._update_channels(channels)
        # make it update also the channels
    def _delete_layout_content(self, layout):
        for i in reversed(range(layout.count())):
            layout.itemAt(i).widget().setParent(None)



    # centralized version of the thing
    def get_image_to_display_including_all_dims(self):
        # print('in get_image_to_display_including_all_dims')
        # returns the image to be displayed --> takes into account all the dimensions at once

        #
        # if self.raw_image is not None:
        #     if i == 0:
        #         self.set_display(self.raw_image)
        #         self.channel = None
        #         # print('original', self.img.metadata)
        #     else:
        #         # print('modified0', self.img.metadata)
        #         # I need a hack when the image is single channel yet I need several masks for it !!!
        #         if self.multichannel_mode and i - 1 >= self.raw_image.shape[-1]:
        #             channel_img = self.raw_image.imCopy(c=0)  # if out of bonds load the first channel
        #         else:
        #             channel_img = self.raw_image.imCopy(c=i - 1)  # it's here that it is affected
        #         self.channel = i - 1
        #         # print('modified1', self.img.metadata)
        #         # print('modified2', channel_img.metadata)
        #         self.set_display(
        #             channel_img)  # maybe do a set display instead rather --> easier to handle --> does a subest of the other
        image_to_display = None
        try:
            # need change just the displayed image
            if has_metadata(self.paint.raw_image) and self.paint.raw_image.metadata['dimensions']:
                # change the respective dim
                # need all the spinner values to be recovered in fact
                # and send the stuff
                # print('dimension exists', self.objectName(),'--> changing it')

                # need gather all the dimensions --> TODO
                dimensions = self.paint.raw_image.metadata['dimensions']
                position_h = dimensions.index('h')
                image_to_display = self.paint.raw_image
                if position_h != 0:
                    # loop for all the dimensions before
                    # print('changing stuff')
                    for pos_dim in range(0, position_h):
                        dim = dimensions[pos_dim]
                        value_to_set = self.get_dim_value_by_name(dim)
                        if value_to_set == None:
                            continue
                        else:
                            image_to_display = image_to_display[value_to_set]
                        # if not dimensions[pos_dim]==self.sender().objectName():
                        #     image_to_display = image_to_display[0]
                        # else:
                        #     image_to_display = image_to_display[self.sender().value()]
                # self.paint.set_display(image_to_display)
                channel_to_display = self.get_selected_channel()

                # if force dimensions --> I need a hack even if
                # hack for GT image editor --> where nb of channels are forced and do not necessarily match the channels of the image --> is that smart to put that here ???
                if channel_to_display is not None and 'c' in dimensions:
                    if self.paint.multichannel_mode and channel_to_display >= self.paint.raw_image.shape[-1]:
                        image_to_display = image_to_display[..., 0]
                    else:
                        image_to_display=image_to_display[...,channel_to_display]
                    # print(image_to_display.shape)
        except:
            traceback.print_exc()
        return image_to_display

    def update_image_dimensions(self):
        self.dimension_sliders = []

        for i in reversed(range(self.dimensions_container.count())):
            # print(i)
            # print(self.dimensions_container.itemAt(i))
            # self.dimensions_container.itemAt(i).widget().setParent(None)
            self._delete_layout_content(self.dimensions_container.itemAt(i))

        if self.paint.raw_image is None:
            return

        # empty the content of the slider and refill it
        try:
            treat_channels_as_a_browsable_dimension = False
            if has_metadata(self.paint.raw_image):
                dimensions = self.paint.raw_image.get_dimensions_as_string()

                # if not self.paint.raw_image.has_c():
                #     nb_of_sliders = len(
                #         self.paint.raw_image.shape) - 2  # or -3 it depends whether I wanna show all the channels at once or not ???
                # else:
                #     nb_of_sliders = len(
                #         self.paint.raw_image.shape) - 2  # or -3 it depends whether I wanna show all the channels at once or not ???

                # print(nb_of_sliders)

                # create an image with plenty of sliders --> TODO
                # make the handling of the image to be displayed directly by the code !!!

                # then I need couple each slider to a dimension

                # dimensions that must have a slider
                # --> all dimensions but hw and maybe c must have a slider --> TODO
                for dim in dimensions:
                    if dim == 'h' or dim == 'y' or dim == 'w' or dim == 'x' or (
                            not treat_channels_as_a_browsable_dimension and dim == 'c'):
                        # we skip dimensions
                        continue
                    # print(dim, 'must have an assoicated slider')

                    # print(self.paint.raw_image.shape, ' toto ', )
                    self.dimensions_container.addLayout(self.create_dim_slider(dimension=dim, max_dim=self.paint.raw_image.shape[dimensions.index(dim)]))
        except:
            traceback.print_exc()



    def set_mask(self, mask):
        self.paint.set_mask(mask)

    def _update_channels(self, img):
        selection = self.channels.currentIndex()
        self.channels.disconnect()
        self.channels.clear()
        comboData = ['merge']
        if img is not None:
            try:
                if img.has_c():
                    for i in range(img.get_dimension('c')):
                        comboData.append(str(i))
            except:
                # assume image has no channel and return None
                # or assume last stuff is channel if image has more than 2
                pass
        # logger.debug('channels found ' + str(comboData))
        self.channels.addItems(comboData)
        # index = self.channels.findData(selection)

        self.channels.currentIndexChanged.connect(self.channelChange)
        # print("data", index)
        if selection != -1 and selection < self.channels.count():
            self.channels.setCurrentIndex(selection)
            # self.channelChange(selection)
        else:
            self.channels.setCurrentIndex(0)
            # self.channelChange(0)


    def freeze(self, bool, level=1):

        self.tb1.setEnabled(True)
        self.tb2.setEnabled(True)
        self.tb3.setEnabled(True)

        if level == 1:
            self.tb2.setEnabled(not bool)
        elif level == 2:
            self.tb2.setEnabled(not bool)
            self.tb3.setEnabled(not bool)

        if bool:
            self.disable_shortcuts()
        else:
            self.enable_shortcuts()
        # remove draw mode maybe and freeze shortcuts

    def create_dim_slider(self, dimension=None, max_dim=1):
        dim_slider_with_label1 = QHBoxLayout()
        label_slider1 = QLabel()
        if dimension is not None:
            label_slider1.setText(dimension)
        fake_dim_slider = QSlider(Qt.Horizontal)
        fake_dim_slider.setMinimum(0)
        fake_dim_slider.setMaximum(max_dim-1)
        dim_slider_with_label1.addWidget(label_slider1)
        dim_slider_with_label1.addWidget(fake_dim_slider)
        # dim_slider_with_label1
        # add a partial?
        fake_dim_slider.valueChanged.connect(self.slider_dimension_value_changed)
        # fake_dim_slider.valueChanged.connect(lambda x: self.delayed_dimension_preview.start(600))
        fake_dim_slider.setObjectName(dimension)
        self.dimension_sliders.append(fake_dim_slider)
        return dim_slider_with_label1

    def get_dim_value_by_name(self, label):
        for slider in self.dimension_sliders:
            if slider.objectName() == label:
                return slider.value()
        return None

    # def get_dims_value_by_name(self, label):
    #     for slider in self.dimension_sliders:
    #         if slider.objectName() == label:
    #             return slider.value()
    #     return None

    # p
    def slider_dimension_value_changed(self):
        # TODO --> do delay only if necessary
        # self.delayed_dimension_preview.stop()
        # self.delayed_dimension_preview.start(600)

        # self.delayed_dimension_preview.timeout.connect

        # if self.delayed_dimension_preview.dis

        # TODO only allow this when the dimension changing is over
        # if the dimension exists --> print it
        # print(self.sender(), self.sender().objectName(), self.sender().value())
        # print(self.sender(), self.sender().objectName())
        # print(sender)
        #
        # try:
        #     # need change just the displayed image
        #     if self.paint.raw_image.metadata['dimensions']:
        #         # change the respective dim
        #         # need all the spinner values to be recovered in fact
        #         # and send the stuff
        #         # print('dimension exists', self.objectName(),'--> changing it')
        #
        #         # need gather all the dimensions --> TODO
        #         dimensions = self.paint.raw_image.metadata['dimensions']
        #         position_h =  dimensions.index('h')
        #         image_to_display = self.paint.raw_image
        #         if position_h!=0:
        #             # loop for all the dimensions before
        #             # print('changing stuff')
        #             for pos_dim in range(0,position_h):
        #                 dim = dimensions[pos_dim]
        #                 value_to_set = self.get_dim_value_by_name(dim)
        #                 if value_to_set == None:
        #                     continue
        #                 else:
        #                     image_to_display = image_to_display[value_to_set]
        #                 # if not dimensions[pos_dim]==self.sender().objectName():
        #                 #     image_to_display = image_to_display[0]
        #                 # else:
        #                 #     image_to_display = image_to_display[self.sender().value()]
        #         self.paint.set_display(image_to_display)
        # except:
        #     traceback.print_exc()

        try:
            meta = None
            try:
                meta = self.paint.raw_image.metadata
            except:
                pass
            # metadata is required to get the luts properly
            self.paint.set_display(self.get_image_to_display_including_all_dims(), metadata=meta)
        except:
            traceback.print_exc()

        # pass

    # then have a single code to handle dynamically all the dimension changes !!! --> TODO


# pas mal faut juste voir ce que je peux faire
# peut etre permetre de mettre des custom panels

if __name__ == '__main__':
    import sys
    from qtpy.QtWidgets import QApplication


    # hand drawing panel
    if False:

        # TODO add a main method so it can be called directly
        # maybe just show a canvas and give it interesting props --> TODO --> really need fix that too!!!


        # should probably have his own scroll bar embedded somewhere

        app = QApplication(sys.argv)


        # quite easy todo in fact

        # no clue why it always works just once


        # maybe make a scrollable TA paint that will handle and behave like that
        # maybe handle also stacks at some point ---> see how
        # --> almost all is ok now
        class overriding_apply(Createpaintwidget):
            # all seems ok now and functions as in TA
            # just do the shift enter to get rid of small cells

            # avec ça ça marche 100% à la TA ... --> cool --> maybe make it a TA drawing pad
            def apply(self):
               self.apply_drawing(minimal_cell_size=0)

            def shift_apply(self):
                # MEGA TODO IMPLEMENT SIZE within this stuff!!!
                self.apply_drawing(minimal_cell_size=10)

            def ctrl_m_apply(self):
                self.manually_reseeded_wshed()

            def save(self):
                self.save_mask()

        w = scrollable_paint(
            custom_paint_panel=overriding_apply())  # ça marche --> permet de mettre des paint panels avec des proprietes particulieres --> assez facile en fait
        # w.set_image('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012.png')
        # w.set_mask('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012/handCorrection.png')
        # w.set_image('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/Optimized_projection_018.png')
        # w.set_mask('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/neo_mask.tif')

        w.set_image('/D/VirtualBox/book_chapter_image_segmentation_final_2021/new_figs/dataset_1/minimal_demo_sample_simplified_bckup_before_messing_with_two_seg_errors/1.tif')
        w.set_mask('/D/VirtualBox/book_chapter_image_segmentation_final_2021/new_figs/dataset_1/minimal_demo_sample_simplified_bckup_before_messing_with_two_seg_errors/1/handCorrection.tif')


        # ça marche mais tt finalizer
        # that seems to work

        # w.freeze(True)
        # w.freeze(False)
        # w.freeze(True)

        # all is so easy to do this way

        # ça marche --> tres facile a tester
        # --> how can I do a zoom
        # ask whether self scroll or not --> if scrollable

        # mask = w.get_mask()
        #
        # plt.imshow(mask)
        # plt.show()

        w.show()
        sys.exit(app.exec_())

    # vectorial panel for drawing ROIs --> that works --> keep this code cause very useful and many bug fix
    # TODO try convert any ROIs to something like a contour using the pavlidis algo --> put this in rps tools

    if True:
        # TODO add a main method so it can be called directly
        # maybe just show a canvas and give it interesting props --> TODO --> really need fix that too!!!

        # should probably have his own scroll bar embedded somewhere

        app = QApplication(sys.argv)


        # quite easy todo in fact

        # no clue why it always works just once

        # maybe make a scrollable TA paint that will handle and behave like that
        # maybe handle also stacks at some point ---> see how
        # --> almost all is ok now
        class overriding_apply(Createpaintwidget):
            # all seems ok now and functions as in TA
            # just do the shift enter to get rid of small cells
            def m_apply(self):
                self.vdp.active = not self.vdp.active
                self.update()

            # avec ça ça marche 100% à la TA ... --> cool --> maybe make it a TA drawing pad
            def apply(self):
                # self.apply_drawing(minimal_cell_size=0)
                pass

            def shift_apply(self):
                # MEGA TODO IMPLEMENT SIZE within this stuff!!!
                # self.apply_drawing(minimal_cell_size=10)
                pass

            def ctrl_m_apply(self):
                # print('seed')
                # self.manually_reseeded_wshed()
                pass

            def save(self):
                # self.save_mask()
                print('saving ROIs')
                print(self.vdp.shapes)
                # create an image the save size as the displayed one and open it
                if self.raw_image is None:
                    return
                if len(self.raw_image.shape)==2:
                    black_image = np.zeros_like(self.raw_image, dtype=np.uint64)
                else:
                    black_image = np.zeros(shape=(*self.raw_image[:-1],), dtype=np.uint64)
                # plt.imshow(black_image)
                # plt.show()
                # make it fill all the ROIs inside that
                im2d = Image2D(black_image)

                for iii,shape in enumerate(self.vdp.shapes):
                    # ideally need clone it and make it the enumerate color +1

                    shape.fill_color = iii+1
                    shape.color = None
                    # print(shape.fill_color)
                    # print(type(shape), shape.fill_color)
                    im2d.annotation.append(shape)



                # im2d.annotation.extend(self.vdp.shapes)

                # need clone the shape

                # I need remove antialiasing
                im = im2d.save(None,quality=QPainter.NonCosmeticDefaultPen)
                im = im2d.convert_qimage_to_numpy(im)

                # restore colors
                for shape in self.vdp.shapes:
                    shape.fill_color = None
                    shape.color = 0xFFFF00

                # faudrait tester en grandeur reelle pr voir si ça marche --> à faire !!!
                if True:
                    # TODO just save the raw file in one channel not in many --> check the  output by the way

                    # from skimage.measure import label
                    # lab = label(im, connectivity=2, background=0)
                    # im=lab

                    # I need remove antialiasing
                    Img(im.astype(np.float64), dimensions='hwc').save('/E/Sample_images/sample_image_neurons_laurence_had/data_test/test_raw.tif')

                    plt.imshow(im)
                    plt.show()

                # what I need to do here now is to save in the folder of the parent à la TA --> can save it as neurons.tif maybe




            # def suppr_pressed(self):
            #     # self.save_mask()
            #     # print('suppr')
            #     self.pa


        test = overriding_apply()
        test.vdp.shape_to_draw = Freehand2D

        # test.vdp.shape_to_draw = Rect2D # ça ça marche --> bug is in the other then...

        # there is a bug in the vectorial but no big deal probably

        w = scrollable_paint(
            custom_paint_panel=test)  # ça marche --> permet de mettre des paint panels avec des proprietes particulieres --> assez facile en fait
        # w.set_image('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012.png')
        # w.set_mask('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012/handCorrection.png')
        # w.set_image('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/Optimized_projection_018.png')
        # w.set_mask('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/neo_mask.tif')

        test.vdp.active = True
        test.vdp.drawing_mode = True

        # QApplication.setOverrideCursor(Qt.WaitCursor)
        # do lengthy process
        # QApplication.restoreOverrideCursor()

        w.set_image('/E/Sample_images/sample_images_FIJI/sample2.lsm')

        # w.set_image('/E/Sample_images/sample_image_neurons_laurence_had/data_test/image.png')
        # w.set_mask('/D/VirtualBox/book_chapter_image_segmentation_final_2021/new_figs/dataset_1/minimal_demo_sample_simplified_bckup_before_messing_with_two_seg_errors/1/handCorrection.tif')

        # w.setCursor(Qt.BlankCursor)

        # ça marche mais tt finalizer
        # that seems to work

        # w.freeze(True)
        # w.freeze(False)
        # w.freeze(True)

        # we force the cursor to be visible --> set it by default
        # cursor = Qt.ArrowCursor
        # w.paint.setCursor(cursor)

        # all is so easy to do this way

        # ça marche --> tres facile a tester
        # --> how can I do a zoom
        # ask whether self scroll or not --> if scrollable

        # mask = w.get_mask()
        #
        # plt.imshow(mask)
        # plt.show()

        w.show()
        # w.setCursor(Qt.BlankCursor)
        sys.exit(app.exec_())



import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import traceback
from timeit import default_timer as time
from qtpy.QtCore import QSize, QTimer
from natsort import natsorted
import glob
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QPushButton, QWidget, QLineEdit, QApplication, QGridLayout, QStyle, QLabel
from epyseg.dialogs.opensave import openFileNameDialog, openDirectoryDialog
import sys
import os

class QLineEditDND(QLineEdit):

    def __init__(self, title, parent, tip_text=None, objectName=''):
        super().__init__(title, parent, objectName=objectName)
        if tip_text is not None:
            self.setPlaceholderText(tip_text)
        self.setAcceptDrops(True)

    def dragEnterEvent(self, e):
        # TODO test on osX to see if that works
        if e.mimeData().hasUrls():
            e.accept()
        else:
            e.ignore()

    # TODO handle several files if needed
    def dropEvent(self, e):
        if e.mimeData().hasUrls():
            self.setText(e.mimeData().urls()[0].toLocalFile())
        # TODO URGENT I guess I should activate the lines below
        #     e.accept()
        # else:
        #     e.ignore()

class OpenFileOrFolderWidget(QWidget):

    def __init__(self, parent_window=None, add_timer_to_changetext=False, show_ok_or_not_icon=False, label_text=None,
                 is_file=False, extensions="All Files (*);;", show_size=False, tip_text=None, objectName='',finalize_text_change_method_overrider=None):
        super().__init__(parent=parent_window)
        # print('changed method',finalize_text_change_method_overrider)
        self.finalize_text_change_method_overrider = finalize_text_change_method_overrider
        self.ok_ico = self.style().standardIcon(QStyle.SP_DialogYesButton).pixmap(QSize(12, 12))
        self.not_ok_ico = self.style().standardIcon(QStyle.SP_DialogNoButton).pixmap(QSize(12, 12))
        self.show_ok_or_not_icon = show_ok_or_not_icon
        self.parent_window = parent_window
        self.add_timer_to_changetext = add_timer_to_changetext
        self.label_text = label_text
        self.is_file = is_file
        self.tip_text = tip_text
        # self.is_file_or_folder = is_file_or_folder
        self.extensions = extensions
        self.show_size = show_size
        self.time_of_last_change = None
        self.initUI(objectName)

    def initUI(self, objectName):
        layout = QGridLayout()
        size_begin = 80
        size_label = 0
        size_end = 20
        size_warning = 0
        size_input_size = 0

        if self.label_text is not None:
            size_label = 10
            size_begin -= size_label

        if self.show_ok_or_not_icon:
            size_warning = 3
            size_begin -= size_warning

        if self.show_size:
            size_input_size = 5
            size_begin -= size_input_size

        layout.setColumnStretch(0, size_label)
        layout.setColumnStretch(1, size_begin)
        layout.setColumnStretch(2, size_end)
        layout.setColumnStretch(3, size_warning)
        layout.setColumnStretch(4, size_input_size)
        layout.setHorizontalSpacing(3)
        layout.setVerticalSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        self.path = QLineEditDND('', self, tip_text=self.tip_text, objectName=objectName)

        if self.add_timer_to_changetext:

            qtimer = QTimer()
            qtimer.setSingleShot(True)
            # self.path.textChanged.connect(lambda: qtimer.start(600))
            self.path.textChanged.connect(lambda x: qtimer.start(600)) # somehow my weird bug is here and only in some conditions
            # # Is there a better way TODO that ????
            # timer.timeout.connect(self._text_changed)
            qtimer.timeout.connect(self.finalize_text_change)
        self.path.setDragEnabled(True)

        open_ico = QIcon.fromTheme("folder-open")
        open_button = QPushButton(open_ico, "Open", self)
        # bt_width = open_button.fontMetrics().boundingRect(open_button.text()).width() + 30
        # open_button.setMaximumWidth(bt_width)
        # if self.is_file_or_folder:
        #     open_button.clicked.connect(self.open_file_or_folder)
        # el
        if self.is_file:
            open_button.clicked.connect(self.open_file)
        else:
            open_button.clicked.connect(self.open_folder)

        self.ok_or_not_ico = QLabel('')
        self.ok_or_not_ico.setPixmap(self.not_ok_ico)

        if self.show_ok_or_not_icon:
            layout.addWidget(self.ok_or_not_ico, 0, 3)

        if self.label_text is not None:
            layout.addWidget(QLabel(self.label_text), 0, 0)

        self.size_label = QLabel('0000')

        if self.show_size:
            layout.addWidget(self.size_label, 0, 4)

        layout.addWidget(self.path, 0, 1)
        layout.addWidget(open_button, 0, 2)
        self.setLayout(layout)

    def open_folder(self):
        if os.path.isdir(self.path.text()):
            self.output_file_or_folder = openDirectoryDialog(parent_window=self.parent_window)
        else:
            self.output_file_or_folder = openDirectoryDialog(parent_window=self.parent_window,
                                                                                 path=self.path.text())
        if self.output_file_or_folder is not None:
            self.path.setText(self.output_file_or_folder)

    def open_file(self):
        if os.path.isfile(self.path.text()):
            self.output_file_or_folder = openFileNameDialog(parent_window=self.parent_window,
                                                                                extensions=self.extensions)
        else:
            self.output_file_or_folder = openFileNameDialog(parent_window=self.parent_window,
                                                                                extensions=self.extensions,
                                                                                path=self.path.text())
        if self.output_file_or_folder is not None:
            self.path.setText(self.output_file_or_folder)

    def set_time_of_last_change(self):
        self.time_of_last_change = time()
        # print('I have been changed', self.time_of_last_change, self) # to debug

    def get_time_of_last_change(self):
        return  self.time_of_last_change

    # def _text_changed(self):
    #     # this stores the time of last change --> can easily determine when the object was last changed
    #     self.set_time_of_last_change()
    #     # self.finalize_text_change()

    def text(self):
        if self.path.text().strip() == '':
            return None
        return self.path.text()

    def finalize_text_change(self):
            # print('finalize_text_change_method_overrider',self.finalize_text_change_method_overrider)
        # try:
            self.set_time_of_last_change()
            if self.finalize_text_change_method_overrider is not None:
                # print('calling neo finalizer')
                # if an overriding method was defined --> use it otherwise use default
                self.finalize_text_change_method_overrider()
            else:
                #
                print('in', self.path.text())
                # pass
        # except:
        #     # traceback.print_stack()
        #     traceback.print_exc()
        #     pass
        # pass

    def set_icon_ok(self, ok):
        if not self.show_ok_or_not_icon:
            return
        if ok:
            self.ok_or_not_ico.setPixmap(self.ok_ico)
        else:
            self.ok_or_not_ico.setPixmap(self.not_ok_ico)

    def set_size(self, size):
        self.size_label.setText(size)

    def get_list_using_glob(self):
        try:
            filenames = [file for file in glob.glob(self.path.text())]
            filenames = natsorted(filenames)  # human-like sorting of file names
            return filenames
        except:
            return None

if __name__ == '__main__':
    # just for a test
    app = QApplication(sys.argv)

    def new_method():
        print('tutu')

    ex = OpenFileOrFolderWidget(parent_window=None, add_timer_to_changetext=True, finalize_text_change_method_overrider=new_method)
    ex.show()
    app.exec_()



import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import traceback
import logging
from itertools import zip_longest
from epyseg.deeplearning.docs.doc2html import markdown_file_to_html, browse_tip
from epyseg.gui.defineROI import DefineROI
from epyseg.utils.loadlist import loadlist
from epyseg.postprocess.gui import PostProcessGUI
from epyseg.uitools.blinker import Blinker
from qtpy.QtWidgets import QDialog, QDialogButtonBox, QPushButton, QToolTip, QHBoxLayout, QFrame
from qtpy.QtWidgets import QWidget, QApplication, QGridLayout, QTabWidget
from qtpy import QtCore
from epyseg.deeplearning.augmentation.generators.data import DataGenerator
from epyseg.gui.open import OpenFileOrFolderWidget
from epyseg.gui.preview import crop_or_preview
from qtpy.QtWidgets import QSpinBox, QComboBox, QVBoxLayout, QLabel, QCheckBox, QRadioButton, QButtonGroup, QGroupBox, \
    QDoubleSpinBox
from qtpy.QtCore import Qt, QPoint
from epyseg.img import Img
import sys
import json
import os
from epyseg.gui.multi_inputs_img import Multiple_inputs
from epyseg.tools.logger import TA_logger # logging

logger = TA_logger()

# TODO set a min size for this stuff so that size fits in smaller windows
class minisel(QDialog):

    def __init__(self, parent_window=None):
        super().__init__(parent=parent_window)
        # d = QDialog()

        # ask whether images are to be used in combination with TA or not
        self.auto_output = QRadioButton('Minimal interface (just use pretrained EPySeg, CARE, ... models)',
                                   objectName='auto_output')
        self.custom_output = QRadioButton('Advanced interface (to train a model or use cutsom parameters for prediction)',
                                     objectName='custom_output')
        predict_output_radio_group = QButtonGroup()
        # predict_output_radio_group.buttonClicked.connect(self.predict_output_mode_changed)

        # default is build a new model
        self.auto_output.setChecked(True)

        # connect radio to output_predictions_to text
        self.auto_output.toggled.connect(self.predict_output_mode_changed)
        self.custom_output.toggled.connect(self.predict_output_mode_changed)

        predict_output_radio_group.addButton(self.auto_output)
        predict_output_radio_group.addButton(self.custom_output)

        # b1 = QPushButton("ok", d)
        # b1.move(50, 50)
        self.setWindowTitle("Dialog")
        # self.setWindowModality(QtCore.Qt.ApplicationModal)

        layout = QGridLayout()

        layout.addWidget(self.auto_output)
        layout.addWidget(self.custom_output)

        line_sep_predict = QFrame()
        line_sep_predict.setFrameShape(QFrame.HLine)
        line_sep_predict.setFrameShadow(QFrame.Sunken)

        layout.addWidget(line_sep_predict)

        self.setLayout(layout)

        # OK and Cancel buttons
        self.buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
                                     QtCore.Qt.Horizontal, self)
        self.buttons.accepted.connect(self. accept)
        self.buttons.rejected.connect(self.reject)
        self.layout().addWidget(self.buttons)

    def get_parameters(self):
        return self.auto_output.isChecked()

    @staticmethod
    def getDataAndParameters(parent_window=None):  #
        # get all the params for augmentation
        dialog = minisel(parent_window=parent_window)
        result = dialog.exec_()
        augment = dialog.get_parameters()
        return (augment, result == QDialog.Accepted)

    def predict_output_mode_changed(self):
        # print('inside')
        # print(self.sender())
        # TODO if the model requires several inputs then update the nb of inputs --> TODO
        pass


if __name__ == '__main__':
    # just for a test

    app = QApplication(sys.argv)
    augment, ok = minisel.getDataAndParameters(parent_window=None)

    print(augment, ok)

    if ok:
        print("continue")
    else:
        print('abort')


    sys.exit(0)



from epyseg.deeplearning.deepl import EZDeepLearning
import os


# TODO --> rename this class

def run_seg(deepTA=None, IS_TA_OUTPUT_MODE = True,    input_channel_of_interest = None, TILE_WIDTH = 256, TILE_HEIGHT = 256, TILE_OVERLAP = 32, INPUT_FOLDER = None, EPYSEG_PRETRAINING = 'Linknet-vgg16-sigmoid-v2', SIZE_FILTER = None, progress_callback=None):
    # predict parameters
    # IS_TA_OUTPUT_MODE = True  # stores as handCorrection.tif in the folder with the same name as the parent file without ext
    # input_channel_of_interest = None  # assumes image is single channel or multichannel nut channel of interest is ch0, needs be changed otherwise, e.g. 1 for channel 1
    # TILE_WIDTH = 256  # 128 # 64
    # TILE_HEIGHT = 256  # 128 # 64
    # TILE_OVERLAP = 32
    # INPUT_FOLDER = '/path/to/files_to_segment/'
    # EPYSEG_PRETRAINING = 'Linknet-vgg16-sigmoid-v2'  # or 'Linknet-vgg16-sigmoid' for v1
    # SIZE_FILTER = None  # 100 # set to 100 to get rid of cells having pixel area < 100 pixels

    # raw code for predict
    if deepTA is None:
        deepTA = EZDeepLearning()
    deepTA.load_or_build(architecture='Linknet', backbone='vgg16', activation='sigmoid', classes=7,
                         pretraining=EPYSEG_PRETRAINING)

    deepTA.get_loaded_model_params()
    deepTA.summary()

    input_shape = deepTA.get_inputs_shape()
    output_shape = deepTA.get_outputs_shape()

    input_normalization = {'method': 'Rescaling (min-max normalization)', 'range': [0, 1], 'individual_channels': True}

    predict_parameters = {}

    predict_parameters["input_channel_of_interest"] = input_channel_of_interest
    predict_parameters["default_input_tile_width"] = TILE_WIDTH
    predict_parameters["default_input_tile_height"] = TILE_HEIGHT
    predict_parameters["default_output_tile_width"] = TILE_WIDTH
    predict_parameters["default_output_tile_height"] = TILE_HEIGHT
    predict_parameters["tile_width_overlap"] = TILE_OVERLAP
    predict_parameters["tile_height_overlap"] = TILE_OVERLAP
    predict_parameters["hq_pred_options"] = "Use all augs (pixel preserving + deteriorating) (Recommended for segmentation)"  # not yet supported on current versions of epyseg
    predict_parameters["post_process_algorithm"] = "default (slow/robust) (epyseg pre-trained model only!)"  # None
    predict_parameters["input_normalization"] = input_normalization
    predict_parameters["filter"] = SIZE_FILTER

    predict_generator = deepTA.get_predict_generator(
        inputs=[INPUT_FOLDER], input_shape=input_shape,
        output_shape=output_shape,
        clip_by_frequency=None, **predict_parameters)  # before was 0.05 which is bad especially with restored images

    if not IS_TA_OUTPUT_MODE:
        predict_output_folder = os.path.join(INPUT_FOLDER, 'predict')
    else:
        predict_output_folder = 'TA_mode'

    deepTA.predict(predict_generator, output_shape, predict_output_folder=predict_output_folder, batch_size=1,progress_callback=progress_callback,
                   **predict_parameters)

    # del deepTA

if __name__ == '__main__':
    INPUT_FOLDER = '/path/to/files_to_segment/'
    IS_TA_OUTPUT_MODE = True  # stores as handCorrection.tif in the folder with the same name as the parent file without ext
    input_channel_of_interest = None  # assumes image is single channel or multichannel nut channel of interest is ch0, needs be changed otherwise, e.g. 1 for channel 1
    TILE_WIDTH = 256  # 128 # 64
    TILE_HEIGHT = 256  # 128 # 64
    TILE_OVERLAP = 32
    EPYSEG_PRETRAINING = 'Linknet-vgg16-sigmoid-v2'  # or 'Linknet-vgg16-sigmoid' for v1
    SIZE_FILTER = None  # 100 # set to 100 to get rid of cells having pixel area < 100 pixels

    run_seg(INPUT_FOLDER = INPUT_FOLDER, IS_TA_OUTPUT_MODE=IS_TA_OUTPUT_MODE, input_channel_of_interest=input_channel_of_interest, TILE_WIDTH=TILE_WIDTH, TILE_HEIGHT=TILE_HEIGHT, TILE_OVERLAP=TILE_OVERLAP,EPYSEG_PRETRAINING=EPYSEG_PRETRAINING, SIZE_FILTER=SIZE_FILTER )


import math


def compute_polarity_nematic(centroid, boundary_coords, image_to_compute_polarity_for):
    center_x = centroid[1]
    center_y = centroid[0]
    # WE MUST SORT OUR PIXELS SO THAT THETA IS MONOTONICALLY INCREASING
    theta_unsorted = []

    for pt in boundary_coords:
        deltaX = pt[1] - center_x
        deltaY = pt[0] - center_y
        rSquared = deltaX * deltaX + deltaY * deltaY
        r = math.sqrt(rSquared)
        cosTheta = deltaX / r
        sinTheta = deltaY / r
        theta = math.atan2(sinTheta, cosTheta)  # BECAUSE OF THE NEED TO CALCULATE dTheta THIS CANNOT BE AVOIDED
        theta_unsorted.append(theta)

    # print('theta_unsorted',theta_unsorted) # so far so good

    #       HashMap<Point, Double> hashmap_point_angle = new HashMap<Point, Double>();
    # #         for (int z = 0; z < boundary_points.size(); z++) {
    # #             hashmap_point_angle.put(boundary_points.get(z), theta_unsorted.get(z));                        //SORT HASHMAP BASED ON ANGLES
    # #         }

    # do I even need that I doubt it strongly
    hashmap_point_angle = {}
    for z in range(len(boundary_coords)):
        # hashmap_point_angle[str(boundary_coords[z])]= theta_unsorted[z]                      #SORT HASHMAP BASED ON ANGLES
        hashmap_point_angle[z] = theta_unsorted[z]  # SORT HASHMAP BASED ON ANGLES
        # alternatively could just store Z

    # print('hashmap_point_angle',hashmap_point_angle)

    # need sort the dict by its values
    cells_sorted_by_theta = {k: v for k, v in sorted(hashmap_point_angle.items(), key=lambda item: item[1])}

    # print('cells_sorted_by_theta', cells_sorted_by_theta) # ça marche mais voir si il faut inverser le truc ou pas

    # COMPUTE NECESSARY TRIGONOMETRIC QUANTITES FOR EACH POINT
    theta_sorted = []
    cos2Theta_sorted = []
    sin2Theta_sorted = []

    for k, v in cells_sorted_by_theta.items():
        pt = boundary_coords[k]
        deltaX = pt[1] - center_x
        deltaY = pt[0] - center_y
        rSquared = deltaX * deltaX + deltaY * deltaY
        r = math.sqrt(rSquared)
        cosTheta = deltaX / r
        sinTheta = deltaY / r
        theta = math.atan2(sinTheta, cosTheta)
        theta_sorted.append(theta)
        cos2Theta_sorted.append(2. * cosTheta * cosTheta - 1.)
        sin2Theta_sorted.append(2. * cosTheta * sinTheta)

    # COMPUTE NEMATIC TENSOR
    # IT IS IMPORTANT THAT theta_sorted BE MONOTONICALLY INCREASING BEFORE THIS LOOP (SORT IT!)

    # print('cos2Theta_sorted',cos2Theta_sorted)
    # print('sin2Theta_sorted',sin2Theta_sorted)

    # compute it for all channels --> gain of time in fact
    polarity = computeS1S2ForChannel(boundary_coords, cells_sorted_by_theta, theta_sorted, cos2Theta_sorted,
                                     sin2Theta_sorted, image_to_compute_polarity_for)

    # print('polarity',polarity)
    return polarity


def computeS1S2ForChannel(boundary_points, cells_sorted_by_theta, theta_sorted, cos2Theta_sorted, sin2Theta_sorted,
                          channel_for_qunantification):
    k = 0
    S1_ch1 = 0
    S2_ch1 = 0
    sint_ch1 = 0.  # sum of intensities

    for idx in cells_sorted_by_theta.keys():
        pt = boundary_points[idx]
        # GET THE CURRENT POINT AND INTENSITY

        curIntensity_ch1 = channel_for_qunantification[pt[0], pt[1]]
        numPoints = len(boundary_points)
        #
        # USE PERIODIC BOUNDARY CONDITIONS
        # theta_last = (k == 0) ? theta_sorted[numPoints - 1] - (math.pi*2.) : theta_sorted[k - 1]
        # theta_next = (k == numPoints - 1) ? theta_sorted[0] + (math.pi*2.) : theta_sorted[k + 1]
        # same in python format:
        theta_last = theta_sorted[numPoints - 1] - (math.pi * 2.) if k == 0 else theta_sorted[k - 1]
        theta_next = theta_sorted[0] + (math.pi * 2.) if k == numPoints - 1 else theta_sorted[k + 1]

        #
        # THIS IS NECESSARY TO GET 0. FOR CONSTANT INTENSITY
        deltaTheta = 0.5 * (theta_next - theta_last)  # COMPUTE deltaTheta AS A THREE-POINT DIFFERENTIAL
        #             //double deltaTheta=theta_sorted.get(j)-theta_next;  //COMPUTE deltaTheta AS A TWO-POINT BACKWARDS DIFFERENTIAL
        #
        # ADD THE SUMMANDS TO THEIR RESPECTIVE SUMS...
        # polarity
        S1_ch1 += cos2Theta_sorted[k] * deltaTheta * curIntensity_ch1
        S2_ch1 += sin2Theta_sorted[k] * deltaTheta * curIntensity_ch1
        sint_ch1 += deltaTheta * curIntensity_ch1
        # //            sint_ch2 += deltaTheta * curIntensity_ch2;
        # //            sint_ch3 += deltaTheta * curIntensity_ch3;
        # //            sint_12bits += deltaTheta * curIntensity_12bits;
        k += 1

    # print(S1_ch1, S2_ch1, sint_ch1)
    return [S1_ch1.tolist(), S2_ch1.tolist(),
            sint_ch1.tolist()]  # this is a nematic object # in fact it does compute for all channels --> it works even better than expected

# --> very good I do have polarity now !!!
# just check that it is the same as in TA and do a nematic object in python too --> because can be useful!!!



import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy.QtCore import QPointF, QLineF, QRectF, QPoint, Qt
from qtpy.QtGui import QBrush, QPen, QColor, QTransform
from math import sqrt

from epyseg.tools.logger import TA_logger

logger = TA_logger()

class Line2D(QLineF):

    def __init__(self, *args, color=0xFFFF00, opacity=1., stroke=0.65, arrow=False, line_style=None, theta=0, **kwargs):
        super(Line2D, self).__init__(*args)
        if not args:
            self.isSet = False
        else:
            self.isSet = True
        self.arrow = arrow
        self.color = color
        self.stroke = stroke
        self.opacity = opacity
        self.scale = 1
        self.translation = QPointF()
        self.line_style = line_style
        # rotation
        self.theta = theta

    def set_rotation(self, theta):
        self.theta = theta

    def set_opacity(self, opacity):
        self.opacity = opacity

    def set_line_style(self,style):
        """
        allows lines to be dashed or dotted or have custom pattern

        :param style: a list of numbers or any of the following Qt.SolidLine, Qt.DashLine, Qt.DashDotLine, Qt.DotLine, Qt.DashDotDotLine but not Qt.CustomDashLine, Qt.CustomDashLine is assumed by default if a list is passed in. None is also a valid value that resets the line --> assume plain line
        :return:
        """
        self.line_style = style
        # if style is a list then assume custom pattern otherwise apply solidline

    def draw(self, painter, draw=True):
        if self.color is None:
            return

        if draw:
            painter.save()

        pen = QPen(QColor(self.color))
        if self.stroke is not None:
            pen.setWidthF(self.stroke)
        if self.line_style is not None:
            if self.line_style in [Qt.SolidLine, Qt.DashLine, Qt.DashDotLine, Qt.DotLine, Qt.DashDotDotLine]:
                pen.setStyle(self.line_style)
            elif isinstance(self.line_style, list):
                pen.setStyle(Qt.CustomDashLine)
                pen.setDashPattern(self.line_style)
        painter.setPen(pen)
        painter.setOpacity(self.opacity)
        if draw:
            # clone the line
            line_to_plot = self.translated(0, 0)
            if self.scale is not None and self.scale != 1:
                p1 = line_to_plot.p1()
                p2 = line_to_plot.p2()
                line_to_plot.setP1(QPointF(p1.x()*self.scale, p1.y()*self.scale))
                line_to_plot.setP2(QPointF(p2.x()*self.scale, p2.y()*self.scale))
            if self.translation is not None:
                line_to_plot.translate(self.translation)
            # print(line_to_plot)
            if self.theta is not None and self.theta != 0:
                painter.translate(line_to_plot.center())
                painter.rotate(self.theta)
                painter.translate(-line_to_plot.center())

            painter.drawLine(line_to_plot)
            painter.restore()
    #
    # def fill(self, painter, draw=True):
    #     if draw:
    #         painter.save()
    #     if self.fill_color is None:
    #         return
    #     painter.setBrush(QBrush(QColor(self.fill_color)))
    #     painter.setOpacity(self.opacity)
    #     if draw:
    #         painter.drawLine(self)
    #         painter.restore()
    #
    # def drawAndFill(self, painter):
    #     painter.save()
    #     self.draw(painter, draw=False)
    #     self.fill(painter, draw=False)
    #     painter.drawLine(self)
    #     painter.restore()

    def contains(self, *args):
        x = 0
        y = 0
        if isinstance(args[0], QPoint) or isinstance(args[0], QPointF):
            x = args[0].x()
            y = args[0].y()
        else:
            x = args[0]
            y = args[1]
        return self.distToSegment(QPointF(x, y), self.p1(), self.p2()) < 10 and self.boundingContains(*args)

    def lineFromPoints(self, x1, y1, x2, y2):
        a = y2 - y1
        b = x1 - x2
        c = a * x1 + b * y1
        return (a, b, c)

    def len(self, v, w):
        return (v.x() - w.x()) ** 2 + (v.y() - w.y()) ** 2

    def distToSegment(self, p, v, w):
        l2 = self.len(v, w)
        if l2 == 0:
            return self.len(p, v)
        t = ((p.x() - v.x()) * (w.x() - v.x()) + (p.y() - v.y()) * (w.y() - v.y())) / l2
        t = max(0, min(1, t))
        return sqrt(self.len(p, QPointF(v.x() + t * (w.x() - v.x()), v.y() + t * (w.y() - v.y()))))

    def boundingContains(self, *args):
        return self.boundingRect().contains(*args)

    # def boundingRect(self, scaled=True):
    #     scale = 1
    #     if not scaled and self.scale is not None:
    #         scale = self.scale
    #     return QRectF(min(self.p1().x(), self.p2().x()) * scale, min(self.p1().y(), self.p2().y()) * scale,
    #                   abs(self.p2().x() - self.p1().x()) * scale, abs(self.p2().y() - self.p1().y()) * scale)
    # TODO handle scale etc
    def boundingRect(self):
        rect = QRectF(min(self.p1().x(), self.p2().x()), min(self.p1().y(), self.p2().y()),
               abs(self.p2().x() - self.p1().x()), abs(self.p2().y() - self.p1().y()))

        try:
            # print('tada')
            if self.theta is not None and self.theta != 0:
                # print('entering')
                center = rect.center()
                # print('entering2')
                t = QTransform().translate(center.x(), center.y()).rotate(self.theta).translate(-center.x(),
                                                                                                -center.y())
               #  print('entering3')
               #  transformed = self.setTransform(t)
               #  print('entering4')
               #  print(transformed)
               #  print(QRectF(min(transformed.p1().x(), transformed.p2().x()), min(transformed.p1().y(), transformed.p2().y()),
               # abs(transformed.p2().x() - transformed.p1().x()), abs(transformed.p2().y() - transformed.p1().y())))
               #  return QRectF(min(transformed.p1().x(), transformed.p2().x()), min(transformed.p1().y(), transformed.p2().y()),
               # abs(transformed.p2().x() - transformed.p1().x()), abs(transformed.p2().y() - transformed.p1().y()))

                # copy.setT
                # print('entering')

                # t = QTransform().translate( center.x(), center.y()).rotate(self.theta).translate(-center.x(), -center.y())
                # # print('entersd')
                transformed = t.map( self)  #// mapRect() returns the bounding rect of the rotated rect

                # print('rotated',rotatedRect )
                # return rotatedRect
                return QRectF(min(transformed.p1().x(), transformed.p2().x()),
                              min(transformed.p1().y(), transformed.p2().y()),
                abs(transformed.p2().x() - transformed.p1().x()), abs(transformed.p2().y() - transformed.p1().y()))
        except:
            pass
        return rect

    def add(self, *args):
        point = args[1]
        self.setP2(point)
        self.isSet = True

    def set_to_scale(self, factor):
        self.scale = factor

    def set_to_translation(self, translation):
        self.translation = translation

    def get_P1(self):
        return self.boundingRect().topLeft()

    # faut pas utiliser ça sinon pbs --> car en fait ce que je veux c'est postionned le point et pas le setter

    def set_P1(self, point):
        current_pos = self.boundingRect().topLeft()
        self.translate(point.x() - current_pos.x(), point.y() - current_pos.y())
        # self.translate(self.translation)
        # if not args:
        #     logger.error("no coordinate set...")
        #     return
        # if len(args) == 1:
        #     self.setP1(args[0])
        # else:
        #     self.setP1(QPointF(args[0], args[1]))

    # def set_P2(self,*args):
    #     if not args:
    #         logger.error("no coordinate set...")
    #         return
    #     if len(args) == 1:
    #         self.setP2(args[0])
    #     else:
    #         self.setP2(QPointF(args[0], args[1]))

    def erode(self, nb_erosion=1):
        self.__computeNewMorphology(sizeChange=-nb_erosion)

    def dilate(self, nb_dilation=1):
        self.__computeNewMorphology(sizeChange=nb_dilation)

    def __computeNewMorphology(self, sizeChange=1):
        currentBoundingRect = self.boundingRect()
        curWidth = currentBoundingRect.width()
        finalWitdth = curWidth + 2. * sizeChange

        if (finalWitdth < 1):
            finalWitdth = 1

        center2D = QPointF(currentBoundingRect.center().x(), currentBoundingRect.center().y())

        scale = finalWitdth / self.boundingRect(scaled=False).width()# divide by original width

        print('new scale', scale)
        self.set_to_scale(scale)

        # need translate according to center otherwise ok

        # self.setCenter(center2D)

if __name__ == '__main__':
    # ça marche --> voici deux examples de shapes
    test = Line2D(0, 0, 100, 100, arrow=True)

    print(test.lineFromPoints(0, 0, 100, 100))
    print(test.contains(0, 0))  # true
    print(test.contains(10, 10))  # true
    print(test.contains(-10, -10))  # false # on line with that equation but outside range
    print(test.contains(0, 18))  # false

    p1 = test.p1()
    print(p1.x(), p1.y())
    p2 = test.p2()
    print(p2.x(), p2.y())
    print(test.arrow)
    print(test.length())  # sqrt 2 --> 141
    # if it's an arrow I can add easily all the stuff I need

    test = Line2D(0, 0, 1, 1)
    p1 = test.p1()
    print(p1.x(), p1.y())
    p2 = test.p2()
    print(p2.x(), p2.y())
    print(test.arrow)
    import math

    print(test.length() == sqrt(2))  # sqrt 2

    test2 = Line2D()
    p1 = test2.p1()
    print(p1.x(), p1.y())
    p2 = test2.p2()
    print(p2.x(), p2.y())
    print(test2.arrow)

    # TODO add preview as an image to gain time --> TODO



# TODO --> add a shift enter to remove small blobs because there are a lot --> can be done as a a post-processing !!!
# pb some channels are kept 0 or 1 when others are attributed 255 --> big pb --> need a fix or need do a post process to recover all the images
# maybe do a check to normalize all the images in the same way ... when I save
# need a small thingy remover and need change the default values
# make it a blob remover
# ok for now all of these things can be done as post process ...


# shortcuts --> Enter --> apply --> needed to ensure edit is taken into account
# C --> increase contrast of displayed image


# that seems to work but I need create the GT in the TA shape --> TODO --> then test it for real

# TODO --> maybe do a version with two lists so that the user can also load the GT even if the GT is not having a generic name ???
# or shall I look in more folder sur as GT ou ground truth ou predict
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
import os.path
import traceback
from functools import partial
from qtpy.QtWidgets import QProgressBar, \
    QLabel, QMessageBox
from qtpy.QtGui import QPixmap, QPainter
from qtpy.QtWidgets import QStackedWidget
from qtpy.QtWidgets import QGridLayout, QPushButton, QFrame
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QApplication
from qtpy import QtWidgets, QtCore, QtGui

from epyseg.draw.shapes.freehand2d import Freehand2D
from epyseg.draw.shapes.image2d import Image2D
from epyseg.img import Img, RGB_to_int24
from epyseg.ta.GUI.paint2 import Createpaintwidget
from epyseg.ta.GUI.scrollablepaint_multichannel_or_GT_masks_editor import scrollablepaint_multichannel_edit
from epyseg.ta.tracking.tools import smart_name_parser
from epyseg.ta.GUI.stackedduallist import dualList
import qtawesome as qta  # icons
from epyseg.ta.GUI.scrollablepaint import scrollable_paint
import numpy as np

# main parameter !!!
nb_of_GT_channels = 10


__MAJOR__ = 0
__MINOR__ = 0
__MICRO__ = 1
__RELEASE__ = 'b'  # https://www.python.org/dev/peps/pep-0440/#public-version-identifiers --> alpha beta, ...
__VERSION__ = ''.join([str(__MAJOR__), '.', str(__MINOR__), '.'.join([str(__MICRO__)]) if __MICRO__ != 0 else '', __RELEASE__])
__AUTHOR__ = 'Benoit Aigouy'
__NAME__ = 'GT editor'
__EMAIL__ = 'baigouy@gmail.com'

DEBUG = False

class GT_editor(QtWidgets.QMainWindow):
    # stop_threads = False

    def __init__(self, parent=None):
        super().__init__(parent)
        self.initUI()

        # global early_stop
        # early_stop = False
        self.setAcceptDrops(True)  # we accept DND but just to transfer it to the current list handlers!!!

    def initUI(self):
        # this is a timer to delay the preview updates upon spin valuechanged
        try:
            screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
            centerPoint = QApplication.desktop().screenGeometry(screen).center()
        except:
            from qtpy.QtGui import QGuiApplication
            centerPoint = QGuiApplication.primaryScreen().geometry().center()

        # should fit in 1024x768 (old computer screens)
        window_width = 1024
        window_height = 700
        self.setGeometry(
            QtCore.QRect(centerPoint.x() - int(window_width / 2), centerPoint.y() - int(window_height / 2),
                         window_width,
                         window_height))

        # set the window icon
        # self.setWindowIcon(QtGui.QIcon('./../../IconsPA/src/main/resources/Icons/ico_packingAnalyzer2.gif'))
        # variantMap =QVariant()
        # variantMap.insert("color", QColor(10, 10, 10));
        self.setWindowIcon(qta.icon('mdi.brain')) # , color=QColor(200, 200, 200)  # very easy to change the color of my icon
        # 'mdi.hexagon-outline'

        # zoom parameters
        self.scale = 1.0
        self.min_scaling_factor = 0.1
        self.max_scaling_factor = 20
        self.zoom_increment = 0.05

        self.setWindowTitle(__NAME__ + ' v' + str(__VERSION__))

        # this is the default paint window of TA
        self.paint = self.create_paint(number_of_GT_channels=nb_of_GT_channels)
        # print(self.paint.paint.vdp.shapes)

        # this is the properties window (allows to set useful parameters that can be used by TA

        self.Stack = QStackedWidget(self)
        self.last_opened_file_name = QLabel('')

        # self.list = QListWidget(self)  # a list that contains files to read or play with
        # self.list.setSelectionMode(QAbstractItemView.ExtendedSelection)
        # self.list.selectionModel().selectionChanged.connect(self.selectionChanged)  # connect it to sel change
        self.list = dualList()
        self.list.setToolTip("Drag and drop your images over this list")
        for lst in self.list.lists:
            lst.list.selectionModel().selectionChanged.connect(self.selectionChanged)
            # print list_connected


        # TODO replace that with napari
        # self.volumeViewerUI()
        self.Stack.addWidget(self.paint)
        # self.Stack.addWidget(self.volumeViewer)

        # create a grid that will contain all the GUI interface
        self.grid = QGridLayout()
        # grid.setSpacing(10)

        # grid.addWidget(self.scrollArea, 0, 0)
        self.grid.addWidget(self.Stack, 0, 0)
        # self.grid.addWidget(self.Stack, 0, 0,2,1)
        self.grid.addWidget(self.list, 0, 1)
        # self.grid.addWidget(self.logger_console, 0, 1)
        # self.grid.addWidget(self.list, 1, 1)
        # The first parameter of the rowStretch method is the row number, the second is the stretch factor. So you need two calls to rowStretch, like this: --> below the first row is occupying 80% and the second 20%
        # self.grid.setRowStretch(0, 75)
        # self.grid.setRowStretch(2, 25)

        # first col 75% second col 25% of total width
        self.grid.setColumnStretch(0, 75)
        self.grid.setColumnStretch(1, 25)

        # void QGridLayout::addLayout(QLayout * layout, int row, int column, int rowSpan, int columnSpan, Qt::Alignment alignment = 0)
        # self.grid.addWidget(self.table_widget, 2, 0, 1, 2)  # spans over one row and 2 columns

        # self.setCentralWidget(self.scrollArea)
        self.setCentralWidget(QFrame())
        self.centralWidget().setLayout(self.grid)

        # self.statusBar().showMessage('Ready')
        statusBar = self.statusBar()  # sets an empty status bar --> then can add messages in it
        self.paint.statusBar = statusBar

        # add progress bar to status bar
        self.pbar = QProgressBar(self)
        self.pbar.setToolTip('Shows current progress')
        self.pbar.setGeometry(200, 80, 250, 20)
        statusBar.addWidget(self.pbar)


        self.about = QPushButton()
        self.about.setIcon(qta.icon('mdi.information-variant', options=[{'scale_factor': 1.5}]))
        self.about.clicked.connect(self.about_dialog)
        self.about.setToolTip('About...')
        statusBar.addWidget(self.about)

        statusBar.addWidget(self.last_opened_file_name)

        # Set up menu bar
        self.mainMenu = self.menuBar()
        self.setMenuBar(self.mainMenu)

        # shall I remove this useless shortcut
        # set drawing window fullscreen
        fullScreenShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_F), self)
        fullScreenShortcut.activated.connect(self.fullScreen)
        fullScreenShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active

        # set drawing window fullscreen
        fullScreenShortcut2 = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_F12), self)
        fullScreenShortcut2.activated.connect(self.fullScreen)
        fullScreenShortcut2.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active

        # exit from full screen TODO add quit the app too ??
        escapeShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Escape), self)
        escapeShortcut.activated.connect(self.escape)
        escapeShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active

        spaceShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Space), self)
        spaceShortcut.activated.connect(self.nextFrame)
        spaceShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active

        backspaceShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Backspace), self)
        backspaceShortcut.activated.connect(self.prevFrame)
        # backspaceShortcut.activated.connect(self.down)
        backspaceShortcut.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active

        #!!!!!!!!!! MEGA IMPORTANT KEEP no clue why not activable --> IN FACT IT WAS NOT ACTIVABLE BECAUSE THE SCROLLABLE PAINT WAS ACTIVATING THIS --> NEED DEACTIVATE IT IN THE PARENT OR HERE IT MUST NOT BE IN THE TWO OF THEM

        # supr = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Delete), self)
        # # backspaceShortcut.activated.connect(self.prevFrame)
        # supr.activated.connect(self.paint.paint.suppr_pressed)
        # supr.setContext(QtCore.Qt.ApplicationShortcut)  # make sure the shorctut always remain active

        # Delete does not work but no clue why ????? mais backspace marche -->
        # supr = QtWidgets.QShortcut(QtGui.QKeySequence(Qt.Key_Delete), self)
        # # supr.activated.connect(self.paint.paint.suppr_pressed)
        # supr.activated.connect(self.down)
        # supr.setContext(QtCore.Qt.ApplicationShortcut)

        # self.blinker = Blinker()
        # self.to_blink_after_worker_execution = None
        # self.threading_enabled = True
        #
        # # self.threads = [] # contains all threads so that I can stop them immediately
        # self.thread = None
        # # self.event_stop = threading.Event() # allows to stop a thread
        # # self.threadpool = QThreadPool()
        # # self.threadpool.setMaxThreadCount(self.threadpool.maxThreadCount() - 1)  # spare one core for system
        #
        # self.overlay = Overlay(self.centralWidget())  # maybe ok if draws please wait somewhere
        # self.overlay.hide()
        #
        # try:
        #     self.oldvalue_CUDA_DEVICE_ORDER = os.environ["CUDA_DEVICE_ORDER"]
        #     # print('ok', self.oldvalue_CUDA_DEVICE_ORDER)
        # except:
        #     self.oldvalue_CUDA_DEVICE_ORDER = ''
        #
        # try:
        #     self.oldvalue_CUDA_VISIBLE_DEVICES = os.environ["CUDA_VISIBLE_DEVICES"]
        #     # print('ok', self.oldvalue_CUDA_VISIBLE_DEVICES)
        # except:
        #     self.oldvalue_CUDA_VISIBLE_DEVICES = ''
        #
        # # button.clicked.connect(self.overlay.show)
        #
        # # self.setAcceptDrops(True)  # KEEP IMPORTANT # --> nb can I delegate the dnd of self to a progeny


    def get_current_TA_path(self):
        selection = self.get_selection()
        if selection is None:
            return None
        return smart_name_parser(selection, ordered_output='TA')

    def get_selection(self):
        selected_tab_idx, _ = self.get_cur_tab_index_and_name()
        self.list.set_list(selected_tab_idx)
        list = self.list.get_list(selected_tab_idx)
        selection = list.get_selection()
        return selection

    def get_selection_index(self):
        selected_tab_idx, _ = self.get_cur_tab_index_and_name()
        self.list.set_list(selected_tab_idx)
        list = self.list.get_list(selected_tab_idx)
        selection = list.get_selection_index()
        return selection

    def get_full_list(self, warn_on_empty_list=True):
        lst = self.list.get_list(self.tabs.currentIndex()).get_full_list()
        if warn_on_empty_list:
            if lst is None or not lst:
                print('Empty list, please load files first')
                # self.to_blink_after_worker_execution = [self.list]
                self.blinker.blink(self.list)
                return
        return lst

    def escape(self):
        if self.Stack.isFullScreen():
            self.fullScreen()

    # brings stuff full screen then restore it back after
    # maybe best is to keep the paint version of full screen ??? --> think about it but ok for now
    def fullScreen(self):
        if not self.Stack.isFullScreen():
            self.Stack.setWindowFlags(
                QtCore.Qt.Window |
                QtCore.Qt.CustomizeWindowHint |
                # QtCore.Qt.WindowTitleHint |
                # QtCore.Qt.WindowCloseButtonHint |
                QtCore.Qt.WindowStaysOnTopHint
            )
            '''
                  self.setWindowFlags(
                    QtCore.Qt.Window |
                    QtCore.Qt.CustomizeWindowHint |
                    QtCore.Qt.WindowTitleHint |
                    QtCore.Qt.WindowCloseButtonHint |
                    QtCore.Qt.WindowStaysOnTopHint
                  )
            '''
            self.Stack.showFullScreen()
        else:
            # settings = QSettings()
            # self.Stack.restoreGeometry(settings.value("geometry")) #.toByteArray()
            self.Stack.setWindowFlags(QtCore.Qt.Widget)
            # self.Stack.setWindowFlags(self.flags)
            self.grid.addWidget(self.Stack, 0, 0)  # pas trop mal mais j'arrive pas à le remettre dans le truc principal
            # dirty hack to make it repaint properly --> obviously not all lines below are required but some are --> need test, the last line is key though
            self.grid.update()
            self.Stack.update()
            self.Stack.show()
            self.centralWidget().setLayout(self.grid)
            self.centralWidget().update()
            self.update()
            self.show()
            self.repaint()
            self.Stack.update()
            self.Stack.repaint()
            self.centralWidget().repaint()

    def down(self):
        print('down')

    def nextFrame(self):
        print('next frame pressed')

    def prevFrame(self):
        print('prev frame pressed')

    def about_dialog(self):
        msg = QMessageBox(parent=self)
        msg.setIcon(QMessageBox.Information)
        msg.setText("TEM segmenter")
        msg.setInformativeText("Copyright 2021-2022\n\nBy Benoit Aigouy\n\nbaigouy@gmail.com") # TODO make it dynamic until the current system year
        msg.setWindowTitle("About...")
        # msg.setDetailedText("The details are as follows:")
        msg.setStandardButtons(QMessageBox.Ok)
        # TODO also add the link to the web page on github for the tuto
        retval = msg.exec_()

    def update_preview_depending_on_selected_tab(self):
        # selected_tab_name = self.tabs.currentWidget()
        # selected_tab_idx = self.tabs.currentIndex()

        # self.list.set_list(0)
        list = self.list.get_list(
            0).list  # NB CHANGE THIS --> NOT SMART TO USE LIST AS A NAME --> MAY CAUSE TROUBLE WITH REAL LIST TYPE OF PYTHON --> call it othrewise !!!!
        # TODO replace by self.get_full_list or something alike
        selected_items = list.selectedItems()
        # selected_tab_name = self.tabs.tabText(selected_tab_idx).lower()

        # store file within the tab --> and no relaod except if forced
        # debug
        # print('selection '+ str(0))

        self.paint.freeze(False)

        if selected_items:
            selected_file = selected_items[0].toolTip()

            # if preview tab is selected need run the update

            # whatever happens I need to draw this image except if already loaded

            # try:
            #     # print
            #       # nb this creates a bug and prevents from real channel selection to use for tracking --> do not show this here or find a trick
            #     if selected_tab_name.endswith('racking'):
            #         # dirty trick to block channels, a better solution is required at some point
            #         self.paint.set_image(None)
            #         self.paint.channels.setCurrentIndex(0)
            #         self.paint.set_image(Img(smart_name_parser(selected_file, 'tracked_cells_resized.tif')))
            #         # self.paint.channels.clear()
            #         # self.paint.channels.addItem('merge')
            #
            #         return
            # except:
            #     # probably tracking image does not exist --> just continue
            #     pass

            # start = timer()
            self.paint.set_image(selected_file)
            # self.statusBar().showMessage('Loading ' + selected_items[0].toolTip()) # hides the bar sadly --> not what i want --> need do better --> add a label to the bar
            self.last_opened_file_name.setText(selected_items[0].toolTip())

            # fixed a seg fault in image --> due to inversion of height and witdh for indexed images
            try:
                # update icon if it does not exist yet #TODO recode that in a better way some day
                if list.currentItem() and list.currentItem().icon().isNull():
                    # logger.debug('Updating icon')
                    icon = QIcon(QPixmap.fromImage(self.paint.paint.image))
                    pixmap = icon.pixmap(24, 24)
                    icon = QIcon(pixmap)
                    list.currentItem().setIcon(icon)
            except:
                # no big deal if can't create a thumb out of the selected image
                traceback.print_exc()
                print('failed creating an icon for the selected file')


            # specific GT magic here for loading masks --> maybe some day rather offer two lists and match them
            # will crash if no file exists --> cannot create a GT...
            if os.path.exists(os.path.join(smart_name_parser(selected_file,'parent'),'predict',smart_name_parser(selected_file,'short'))):
                # seach for typical GT stuff produced by epyseg...

                # print('this is the loaded mask',os.path.join(smart_name_parser(selected_file,'parent'),'predict',smart_name_parser(selected_file,'short'))) # l'image chargée est bonne mais pb comment faire pr avoir ts les masks en 3D vraiment appliqués correctement ???


                self.paint.set_mask(os.path.join(smart_name_parser(selected_file,'parent'),'predict',smart_name_parser(selected_file,'short')))
            elif os.path.exists(os.path.join(smart_name_parser(selected_file,'parent'),'GT',smart_name_parser(selected_file,'short'))):
                # search in GT folder...
                self.paint.set_mask(os.path.join(smart_name_parser(selected_file,'parent'),'GT',smart_name_parser(selected_file,'short')))
            elif os.path.exists(os.path.join(smart_name_parser(selected_file,'parent'),'ground_truth',smart_name_parser(selected_file,'short'))):
                self.paint.set_mask(os.path.join(smart_name_parser(selected_file,'parent'),'ground_truth',smart_name_parser(selected_file,'short')))
            elif os.path.exists(os.path.join(smart_name_parser(smart_name_parser(selected_file,'parent'),'parent'),'GT',smart_name_parser(selected_file,'short'))):
                # search in GT folder... in parent of parent orga org GT
                self.paint.set_mask(os.path.join(smart_name_parser(smart_name_parser(selected_file,'parent'),'parent'),'GT',smart_name_parser(selected_file,'short')))
            elif os.path.exists(os.path.join(smart_name_parser(smart_name_parser(selected_file,'parent'),'parent'),'ground_truth',smart_name_parser(selected_file,'short'))):
                self.paint.set_mask(os.path.join(smart_name_parser(smart_name_parser(selected_file,'parent'),'parent'),'ground_truth',smart_name_parser(selected_file,'short')))
            elif os.path.exists(smart_name_parser(selected_file,'training_3_classes.tif')):
                self.paint.set_mask(smart_name_parser(selected_file,'training_3_classes.tif'))
            elif os.path.exists(smart_name_parser(selected_file, 'handCorrection.tif')):
                self.paint.set_mask(smart_name_parser(selected_file, 'handCorrection.tif'))
            else:
                # image not found --> need create a fake empty image --> not sure what I do below is the smartest but maybe ok still
                try:
                    # if len(self.paint.paint.raw_image.shape)<=2:
                    #     self.paint.set_mask(np.zeros(shape=(*self.paint.paint.raw_image.shape,),dtype=np.uint8))

                        # print(self.paint.paint.raw_image.shape)
                        self.paint.set_mask(np.zeros(shape=(self.paint.paint.raw_image.shape[0],self.paint.paint.raw_image.shape[1]),dtype=np.uint8))
                    # else:
                    #     self.paint.set_mask(np.zeros(shape=(*self.paint.paint.raw_image.shape[0:-1], 1), dtype=np.uint8))
                except:
                    traceback.print_exc()
                # self.paint.set_mask(None)

            # # try load the axon mask
            # try:
            #     self.paint.paint.vdp.shapes = []
            #     path_to_axons = smart_name_parser(selected_file,'axons.tif')
            #     if os.path.exists(path_to_axons):
            #         axons = Img(path_to_axons)
            #
            #         contours = get_contours(axons)
            #         if contours:
            #             for contour in contours:
            #                 # color = 0xFF0000, stroke = 2.5)
            #                 contour.color = 0xFF0000
            #                 contour.stroke = 2.5
            #             self.paint.paint.vdp.shapes.extend(contours)
            # except:
            #     traceback.print_exc()
            #     # self.paint.paint.vdp.shapes = []

            # update content depending on tabs in fact
            # if selected_tab_name.startswith('seg'):
            #
            #     # enable mask and drawing
            #     self.paint.maskVisible = True
            #     self.paint.enableMouseTracking()
            #     TA_path_alternative, TA_path = smart_name_parser(
            #         selected_file,
            #         ordered_output=['handCorrection.png', 'handCorrection.tif'])
            #     mask_name = None
            #     if os.path.isfile(TA_path):
            #         mask_name = TA_path
            #     else:
            #         if os.path.isfile(TA_path_alternative):
            #             mask_name = TA_path_alternative
            #     if mask_name is not None:
            #         self.paint.set_mask(mask_name)
            # else:
            #     # disable mask and drawing
            #
            #     self.paint.maskVisible = False
            #     self.paint.disableMouseTracking()
        else:
            self.paint.set_image(None)
            self.last_opened_file_name.setText('')

    def selectionChanged(self):
        try:
            self.update_preview_depending_on_selected_tab()
            # print('TODO --> DO CODE')
        except:
            traceback.print_exc()


    def create_paint(self, number_of_GT_channels=10):

        w = scrollablepaint_multichannel_edit(force_nb_of_channels=number_of_GT_channels)
        w.paint.multichannel_mode = True
        # w.paint.drawing_enabled = True
        # w.enableMouseTracking()
        # w.paint.setMouseTracking(True)

        # w = scrollable_paint(custom_paint_panel=test)

        # test.vdp.active = True
        # test.vdp.drawing_mode = True

        # w.set_image(
        #     '/E/Sample_images/sample_image_neurons_laurence_had/data_test/image.png')

        return w

    def dragEnterEvent(self, event):
        # we transfer all the DNDs to the list
        # selected_tab_idx, _ = self.get_cur_tab_index_and_name()
        self.list.get_list(0).dragEnterEvent(event)

    def dragMoveEvent(self, event):
        # we transfer all the DNDs to the list
        # selected_tab_idx, _ = self.get_cur_tab_index_and_name()
        self.list.get_list(0).dragMoveEvent(event)

    def dropEvent(self, event):
        # we transfer all the DNDs to the list
        # selected_tab_idx, _ = self.get_cur_tab_index_and_name()
        self.list.get_list(0).dropEvent(event)


if __name__ == '__main__':
    import sys



    # spliuiting a 3D image into 2D images would roughly work --> I just need see the most efficient way of doing that --> keep that for later but really almost all is ok now



    # somehow this tool assumes mask must necessarily be 2D and that is not necessarily the case --> it could be useful to have a tool to edit full 3D masks --> maybe do a specific tool just for that --> even different from the stuff
    # think about that
    # tester si j'arrive à segmenter des images assez complexes et full 3D --> à tester en fait mais peut peut etre marcher


    # this tool really does not work for 3D images --> it saves and loads masks wrongly --> I need a really hard hack to get it to work -->
    # maybe ok for now as i can split the image and edit the sequence as a 2D stuff
    # and then reconstitute it in the very end --> TODO


    app = QApplication(sys.argv)
    # set app icon
    # app_icon = QtGui.QIcon('./../../IconsPA/src/main/resources/Icons/ico_packingAnalyzer2.gif')  # add taskbar icon
    # app_icon = qta.icon('mdi.hexagon-multiple-outline', color=QColor(200, 200, 200))
    # app_icon.addFile('gui/icons/16x16.png', QtCore.QSize(16, 16))
    # app_icon.addFile('gui/icons/24x24.png', QtCore.QSize(24, 24))
    # app_icon.addFile('gui/icons/32x32.png', QtCore.QSize(32, 32))
    # app_icon.addFile('gui/icons/48x48.png', QtCore.QSize(48, 48))
    # app_icon.addFile('gui/icons/256x256.png', QtCore.QSize(256, 256))
    # app.setWindowIcon(app_icon)
    # QApplication.setWindowIcon(app_icon)
    w = GT_editor()
    # w.setWindowIcon(app_icon)
    w.show()
    sys.exit(app.exec_())


# channels seem be fixed --> cool
# finalize whsed and corrections --> make sure they are channel dependent
# check which channels are being restored by the new wshed within the paint arena
import matplotlib.pyplot as plt
import numpy as np
import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import QSize, Qt
from qtpy.QtGui import QPalette, QKeySequence, QPainter
from qtpy.QtWidgets import QScrollArea, QVBoxLayout, QWidget, QSpinBox, QComboBox, QToolBar, QStatusBar, QLabel, \
    QHBoxLayout, QAction
import qtawesome as qta

from epyseg.draw.shapes.freehand2d import Freehand2D
from epyseg.draw.shapes.image2d import Image2D
from epyseg.draw.shapes.rect2d import Rect2D
from epyseg.img import Img, int24_to_RGB
from epyseg.ta.GUI.paint2 import Createpaintwidget

# TODO --> could maybe add icons just below scroll if needed to add actions
# j'arrive pas à creer un truc dejà self embedded
# make it also handle other stuff

# --> nb replace the Createpaintwidget in multi_image_display.py by this --> better in fact
# shall I do full screen ???
# TODO handle channels too if needed
# could store mask and memory of drawing here too --> maybe on mouse release ???
# avec des ctrl+Z et des ctrl+R


# need zooms
#

class scrollable_paint(QWidget):

    def __init__(self, custom_paint_panel=None):
        super().__init__()
        if custom_paint_panel is None:
            self.paint = Createpaintwidget()
        else:
            self.paint = custom_paint_panel
        layout = QVBoxLayout()
        self.is_width_for_alternating = True
        # same as the other with embed in scroll area
        self.scrollArea = QScrollArea()
        self.scrollArea.setBackgroundRole(QPalette.Dark)
        self.scrollArea.setWidget(self.paint)
        self.paint.scrollArea = self.scrollArea
        layout.addWidget(self.scrollArea)
        self.setLayout(layout)
        status_bar = QStatusBar()
        layout.addWidget(status_bar)
        self.paint.statusBar = status_bar
        self.add_shortcuts()
        self.drawing_commands()

    # this guy should handle the shortcuts it can and pass the rest to the other layer
    def add_shortcuts(self):
        # print('SHORTCUTS ADDED')

        # if True:
        #     return
        zoomPlus = QtWidgets.QShortcut("Ctrl+Shift+=", self)
        zoomPlus.activated.connect(self.zoomIn)
        zoomPlus.setContext(QtCore.Qt.ApplicationShortcut)



        # that works and disconnects all
        # zoomPlus.disconnect()



        zoomPlus2 = QtWidgets.QShortcut("Ctrl++", self)
        zoomPlus2.activated.connect(self.zoomIn)
        zoomPlus2.setContext(QtCore.Qt.ApplicationShortcut)

        zoomMinus = QtWidgets.QShortcut("Ctrl+Shift+-", self)
        zoomMinus.activated.connect(self.zoomOut)
        zoomMinus.setContext(QtCore.Qt.ApplicationShortcut)

        zoomMinus2 = QtWidgets.QShortcut("Ctrl+-", self)
        zoomMinus2.activated.connect(self.zoomOut)
        zoomMinus2.setContext(QtCore.Qt.ApplicationShortcut)

        ctrl0 = QtWidgets.QShortcut("Ctrl+0", self)
        ctrl0.activated.connect(self.zoom_reset)
        ctrl0.setContext(QtCore.Qt.ApplicationShortcut)

        self.ctrlS = QtWidgets.QShortcut("Ctrl+S", self)
        self.ctrlS.activated.connect(self.paint.save)
        self.ctrlS.setContext(QtCore.Qt.ApplicationShortcut)


        self.ctrlM = QtWidgets.QShortcut("Ctrl+M", self)
        self.ctrlM.activated.connect(self.paint.ctrl_m_apply)
        self.ctrlM.setContext(QtCore.Qt.ApplicationShortcut)

        self.shrtM = QtWidgets.QShortcut("M", self)
        self.shrtM.activated.connect(self.paint.m_apply)
        self.shrtM.setContext(QtCore.Qt.ApplicationShortcut)

        self.increase_contrastC = QtWidgets.QShortcut('C', self)
        self.increase_contrastC.activated.connect(self.paint.increase_contrast)
        self.increase_contrastC.setContext(QtCore.Qt.ApplicationShortcut)

        # I can connect to progeny --> makes sense and really easy --> cool
        self.enterShortcut = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Return), self)
        self.enterShortcut.activated.connect(self.paint.apply)
        self.enterShortcut.setContext(QtCore.Qt.ApplicationShortcut)

        self.enterShortcut2 = QtWidgets.QShortcut(QtGui.QKeySequence(QtCore.Qt.Key_Enter), self)
        self.enterShortcut2.activated.connect(self.paint.apply)
        self.enterShortcut2.setContext(QtCore.Qt.ApplicationShortcut)

        # en fait je peux aussi connecter des trucs au descendant --> pas bete et tres facile à faire!!!

        # pkoi ça ne marche pas !!!
        self.shiftEnterShortcut = QtWidgets.QShortcut("Shift+Enter", self)
        self.shiftEnterShortcut.activated.connect(self.paint.shift_apply)
        self.shiftEnterShortcut.setContext(QtCore.Qt.ApplicationShortcut)
        #
        self.shiftEnterShortcut2 = QtWidgets.QShortcut("Shift+Return", self)
        self.shiftEnterShortcut2.activated.connect(self.paint.shift_apply)
        self.shiftEnterShortcut2.setContext(QtCore.Qt.ApplicationShortcut)

        self.ctrl_shift_S_grab_screen_shot = QtWidgets.QShortcut('Ctrl+Shift+S', self)
        self.ctrl_shift_S_grab_screen_shot.activated.connect(self.paint.grab_screen_shot)
        self.ctrl_shift_S_grab_screen_shot.setContext(QtCore.Qt.ApplicationShortcut)

        self.supr = QtWidgets.QShortcut(QKeySequence(Qt.Key_Delete), self)
        self.supr.activated.connect(self.paint.suppr_pressed)
        self.supr.setContext(QtCore.Qt.ApplicationShortcut)



        # peut etre faire un fullscreen shortcut --> can be useful

    def enable_shortcuts(self):
        self.disable_shortcuts()
        self.ctrlS.activated.connect(self.paint.save)
        self.ctrlM.activated.connect(self.paint.ctrl_m_apply)
        self.shrtM.activated.connect(self.paint.m_apply)
        self.enterShortcut.activated.connect(self.paint.apply)
        self.enterShortcut2.activated.connect(self.paint.apply)
        self.shiftEnterShortcut.activated.connect(self.paint.shift_apply)
        self.shiftEnterShortcut2.activated.connect(self.paint.shift_apply)

    def disable_shortcuts(self):
        # TODO
        try:
            self.ctrlS.disconnect()
        except:
            pass
        # TODO
        try:
            self.ctrlM.disconnect()
        except:
            pass
        try:
            self.shrtM.disconnect()
        except:
            pass
        try:
            self.enterShortcut.disconnect()
        except:
            pass
        try:
            self.enterShortcut2.disconnect()
        except:
            pass
        try:
            self.shiftEnterShortcut.disconnect()
        except:
            pass
        try:
            self.shiftEnterShortcut2.disconnect()
        except:
            pass

    def enableMouseTracking(self):
        self.paint.drawing_enabled = True

    def disableMouseTracking(self):
        self.paint.drawing_enabled = False



    # ça marche --> ajouter fit to width ou fit to height
    def fit_to_width_or_height(self):
        if self.is_width_for_alternating:
            self.fit_to_width()
        else:
            self.fit_to_height()
        self.is_width_for_alternating = not self.is_width_for_alternating

    # almost there but small bugs
    def fit_to_width(self):
        if self.paint.image is None:
            return
        # self.scrollArea = self.scroll_areas[nb]
        width = self.scrollArea.width() - 2
        width -= self.scrollArea.verticalScrollBar().sizeHint().width()
        # height-=self.scrollArea.horizontalScrollBar().sizeHint().height()
        width_im = self.paint.image.width()
        scale = width / width_im
        self.scaleImage(scale)

    def fit_to_height(self, nb=0):
        # paint = self.paints[nb]
        if self.paint.image is None:
            return
        # scrollArea = self.scroll_areas[nb]
        height = self.scrollArea.height() - 2
        height -= self.scrollArea.horizontalScrollBar().sizeHint().height()
        height_im = self.paint.image.height()
        scale = height / height_im
        self.scaleImage(scale)

    def fit_to_window(self):
        # compute best fit
        # just get the size of the stuff and best fit it
        # paint = self.paints[nb]
        if self.paint.image is None:
            return

        # QScrollArea.ensureWidgetVisible(
        # QScrollArea.ensureVisible
        # QScrollArea.setWidgetResizable
        # scrollArea = self.scroll_areas[nb]
        width = self.scrollArea.width() - 2  # required to make sure bars not visible
        height =self. scrollArea.height() - 2

        # scale image that it fits in --> change scale

        # width-=self.scrollArea.verticalScrollBar().sizeHint().width()
        # height-=self.scrollArea.horizontalScrollBar().sizeHint().height()

        height_im = self.paint.image.height()
        width_im = self.paint.image.width()
        scale = height / height_im
        if width / width_im < scale:
            scale = width / width_im
        self.scaleImage(scale)




    def zoomIn(self):
        # 10% more each time
        self.scaleImage(self.paint.scale + (self.paint.scale*10.0/100.0))# une hausse de 10% à chaque fois

    def zoomOut(self):
        # 10% less each time
        self.scaleImage(self.paint.scale - (self.paint.scale*10.0/100.0))

    def zoom_reset(self):
        # resets zoom
        self.scaleImage(1.0)

    # pb this is redundant with the main GUI and not in sync --> can cause a lot of trouble --> dactivate for now
    def full_screen(self):
            if not self.isFullScreen():
                self.fullscreen.setIcon(qta.icon('mdi.fullscreen-exit'))
                self.fullscreen.setToolTip('Exit full screen')
                self.showFullScreen()
                # self.setWindowFlags(
                #     QtCore.Qt.Window |
                #     QtCore.Qt.CustomizeWindowHint |
                #     # QtCore.Qt.WindowTitleHint |
                #     # QtCore.Qt.WindowCloseButtonHint |
                #     QtCore.Qt.WindowStaysOnTopHint
                # )
            else:
                # settings = QSettings()
                # self.Stack.restoreGeometry(settings.value("geometry")) #.toByteArray()
                self.fullscreen.setIcon(qta.icon('mdi.fullscreen'))
                self.fullscreen.setToolTip('Enter full screen')
                self.setWindowFlags(QtCore.Qt.Widget)
                # # self.Stack.setWindowFlags(self.flags)
                # self.grid.addWidget(self.Stack, 0,
                #                     0)  # pas trop mal mais j'arrive pas à le remettre dans le truc principal
                # # dirty hack to make it repaint properly --> obviously not all lines below are required but some are --> need test, the last line is key though
                # self.grid.update()
                # self.Stack.update()
                # self.Stack.show()
                # self.centralWidget().setLayout(self.grid)
                # self.centralWidget().update()
                # self.update()
                self.show()
                self.repaint()
                # self.Stack.update()
                # self.Stack.repaint()
                # self.centralWidget().repaint()

    def scaleImage(self, scale):
        self.paint.set_scale(scale)
        if self.paint.image is not None:
            self.paint.resize(self.paint.scale * self.paint.image.size())
        else:
            self.paint.resize(QSize(0, 0))
        self.paint.update()

    # can maybe add pen stuff here --> simpler and easier
    # need zoom buttons here too
    '''
    buttons/icons to add at some point
    
        
        "Cell smaller than this value will be destroyed"
        "small_cell_remover"
        
        
        "/Icons/refresh.png"
        "run 2 seeds watershed (click twice in a poorly segmented cell to have a the bond be created using the watershed algorithm)"
        
        "/Icons/Knob Snapback.png"
        "Undo"
        
        "/Icons/Knob Valid Blue.png"
        "remove small cells and apply correction"
        
        "/Icons/Knob Valid Green.png"
        "apply correction"
        
        "/Icons/channels.png"
        "Channel Selection"
        "Channel to use for Watershed, polarity, recentering,...")
        
        "/Icons/onebit_11.png"
        "Save"
        
        
        "/Icons/eye.png"
        "View below the mask (lasts 2s)"
        
        
        "/Icons/zoom in.png"
        "Zoom +"
        
        "/Icons/zoom out.png"
        "Zoom -"
        
        "/Icons/1in1.png"
        "1:1 aspect ratio"
        
        "/Icons/fit_2_screen.gif"
        "Fit to panel"
        
        
        "/Icons/double_diagonal_arrow.gif"
        "Alternate between best fit in height or in width"
        
        
        "/Icons/Import Picture Document.png"
        "Send Current View To The System Clipboard"
        
        "/Icons/pipet_small.png"
        "Color pipet"
        
        "Mask col"
        "mask color"
        
        "Correct Drift" 
    
    '''

    # add shortcut for show hide mask and maybe also full screen --> not a good idea

    def drawing_commands(self):
        self.penSize = QSpinBox(objectName='penSize')
        self.penSize.setToolTip('Change this value to change pen size')
        self.penSize.setSingleStep(1)
        self.penSize.setRange(1, 256)
        self.penSize.setValue(3)
        self.penSize.valueChanged.connect(self.penSizechange)

        self.channels = QComboBox(objectName='channels')
        self.channels.setToolTip('Select the channel of interest')
        self.channels.addItem("merge")
        # self.channels.addItem("0")
        # self.channels.addItems(["1", "2", "3"])
        self.channels.currentIndexChanged.connect(self.channelChange)


        hlayout_of_toolbars = QHBoxLayout()
        hlayout_of_toolbars.setAlignment(Qt.AlignLeft)
        self.tb1 = QToolBar()
        self.tb2 = QToolBar()
        self.tb3 = QToolBar()


        pencil_action = QAction(qta.icon('ei.pencil'), 'Pen Size', self)
        pencil_action.setEnabled(False)
        self.tb2.addAction(pencil_action)
        # tb.addAction("sq...")
        self.tb2.addWidget(self.penSize)

        # toolButton = QToolButton()
        # toolButton.setText("Draw")
        # toolButton.setIcon(self.style().standardIcon(QStyle.SP_DriveFDIcon))
        # tb.addWidget(toolButton)

        # fa5.save
        save_action = QAction(qta.icon('fa5.save'), 'Save mask', self)
        save_action.triggered.connect(self.paint.save)
        self.tb2.addAction(save_action)


        # toolButton2 = QToolButton()
        # toolButton2.setText("Draw")
        # toolButton2.setIcon(self.style().standardIcon(QStyle.SP_DialogApplyButton))
        # tb.addWidget(toolButton2)


        show_hide_mask =QAction(qta.icon('fa.bullseye'), 'Show/hide mask (same as pressing "M")', self)
        show_hide_mask.triggered.connect(self.paint.m_apply)
        self.tb2.addAction(show_hide_mask)

        apply_drawing = QAction(qta.icon('fa.check'), 'Apply drawing (same as pressing the "Enter" key)', self)
        apply_drawing.triggered.connect(self.paint.apply)
        self.tb2.addAction(apply_drawing)

        # toolButton3 = QToolButton()
        # toolButton3.setText("Draw")
        # toolButton3.setIcon(self.style().standardIcon(QStyle.SP_DialogCancelButton))
        # tb.addWidget(toolButton3)

        # toolButton4 = QToolButton()
        # toolButton4.setText("Draw")
        # toolButton4.setIcon(self.style().standardIcon(QStyle.SP_DialogCloseButton))
        # tb.addWidget(toolButton4)

        # toolButton5 = QToolButton()
        # toolButton5.setText("Draw")
        # toolButton5.setIcon(self.style().standardIcon(QStyle.SP_TrashIcon))
        # tb.addWidget(toolButton5)

        # toolButton6 = QToolButton()
        # toolButton6.setText("Draw")
        # toolButton6.setIcon(self.style().standardIcon(QStyle.SP_ArrowBack))
        # tb.addWidget(toolButton6)

        # toolButton7 = QToolButton()
        # toolButton7.setText("Draw")
        # toolButton7.setIcon(self.style().standardIcon(QStyle.SP_BrowserReload))
        # tb.addWidget(toolButton7)

        self.small_cell_size = QSpinBox()
        self.small_cell_size.setToolTip('Select the minimal pixel area for a cell to be considered as a cell.\nAny cell having an area less than this value will be removed.')
        self.small_cell_size.setSingleStep(1)
        self.small_cell_size.setRange(1, 1000000)
        self.small_cell_size.setValue(10)
        self.small_cell_size.valueChanged.connect(self.small_cell_size_changed)

        self.tb2.addWidget(self.small_cell_size)
        apply_rm_small_cells = QAction(qta.icon('mdi.check-underline'), 'Apply drawing and remove small cells (same as pressing "Shift+Enter")', self)
        apply_rm_small_cells.triggered.connect(self.paint.shift_apply)# self.paint.shift_apply
        self.tb2.addAction(apply_rm_small_cells)





        # marche pas --> comment faire pr que ça marche...
        local_reshed = QAction(qta.icon('mdi.reload'), 'Locally seeded watershed (same as pressing "Ctrl/Cmd + M")', self)
        local_reshed.triggered.connect(self.paint.ctrl_m_apply)# self.paint.shift_apply
        self.tb2.addAction(local_reshed)
        # hgjhghj

        # toolButton8 = QToolButton()
        # toolButton8.setText("Zoom+")
        # toolButton8.setIcon(qta.icon('ei.zoom-in')) # mdi.magnify-plus-outline --> marche aussi
        # tb.addWidget(toolButton8)

        # toolButton9 = QToolButton()
        # toolButton9.setText("Zoom-")
        # toolButton9.setIcon(qta.icon('ei.zoom-out'))
        # tb.addWidget(toolButton9)

        zoom_plus = QAction(qta.icon('ei.zoom-in'), 'Zoom+', self)
        zoom_plus.triggered.connect(self.zoomIn)
        self.tb1.addAction(zoom_plus)

        zoom_minus = QAction(qta.icon('ei.zoom-out'), 'Zoom-', self)
        zoom_minus.triggered.connect(self.zoomOut)
        self.tb1.addAction(zoom_minus)


        zoom_width_or_height = QAction(qta.icon('mdi.fit-to-page-outline'), 'Alternates between best fit in width and height', self)
        zoom_width_or_height.triggered.connect(self.fit_to_width_or_height)
        self.tb1.addAction(zoom_width_or_height)

        zoom_0 = QAction(qta.icon('ei.resize-full'), 'Reset zoom', self)
        zoom_0.triggered.connect(self.zoom_reset)
        self.tb1.addAction(zoom_0)

        # redundant with TA full screen and not in sync --> disable for now
        # self.fullscreen = QtWidgets.QAction(qta.icon('mdi.fullscreen'), 'Full screen', self)
        # self.fullscreen.setToolTip('Enter full screen')
        # self.fullscreen.triggered.connect(self.full_screen)
        # self.tb1.addAction(self.fullscreen)




        # exit_fullscreen = QtWidgets.QAction(qta.icon('mdi.fullscreen'), 'Full screen', self)
        # exit_fullscreen.triggered.connect(self.showFullScreen)
        # self.tb1.addAction(fullscreen)



        self.channel_label = QLabel('Channels:')
        self.channel_label.setToolTip('Select the channel of interest')
        self.tb3.addWidget(self.channel_label)
        self.tb3.addWidget(self.channels)

        # mdi.fullscreen
        # mdi.fullscreen-exit
        # ei.fullscreen --> a bit less good I find
        # fa.bullseye

        # could have show and hide mask here --> see how I can do that --> but do not delete it in fact

        # self.layout().addWidget(self.tb1)
        hlayout_of_toolbars.addWidget(self.tb2)


        # self.tb2.addAction('toto')
        # self.tb2.addAction('ttutu')
        # self.tb2.addAction('tata')
        hlayout_of_toolbars.addWidget(self.tb3)


        # one for the zooms
        # one for drawing and apply different things
        # TODO do several toolbars to deactivate different levels of stuff
        # and only allow save in the appropriate stuff

        # self.tb3.addAction('toto')
        # self.tb3.addAction('ttutu')
        # self.tb3.addAction('tata')
        hlayout_of_toolbars.addWidget(self.tb1)

        self.layout().addLayout(hlayout_of_toolbars)

    def get_selected_channel(self):
        idx = self.channels.currentIndex()
        if idx == 0:
            return None
        else:
            return idx-1

    # TODO also implement channel change directly within the display tool
    # if merge is applied --> apply on average of all channels --> maybe not so smart an idea but ok to start with and better than what I do in TA
    def channelChange(self, i):
        # update displayed image depending on channel
        # dqqsdqsdqsd
        # pass
        # try change channel if
        self.paint.channelChange(i)
        # print('in channel change !!!')
        # if self.img is not None:
        #     # print('in', self.img.metadata)
        #     if self.Stack.currentIndex() == 0:
        #         # need copy the image --> implement that
        #         # print(self.img[..., i].copy())
        #         # print(self.img[..., i])
        #         if i == 0:
        #             self.paint.setImage(self.img)
        #             # print('original', self.img.metadata)
        #         else:
        #             # print('modified0', self.img.metadata)
        #             channel_img = self.img.imCopy(c=i - 1)  # it's here that it is affected
        #             # print('modified1', self.img.metadata)
        #             # print('modified2', channel_img.metadata)
        #             self.paint.setImage(channel_img)
        #         self.paint.update()
        # else:
        #     # logger.error("Not implemented yet TODO add support for channels in 3D viewer")
        #     # sdqdqsdsqdqsd
        #     self.loadVolume()
        # or reimplement that for multi channels !!!



    # TODO maybe handle pen size change directly within the display --> can be useful by the way
    def small_cell_size_changed(self):
        self.paint.minimal_cell_size = self.small_cell_size.value()

    # TODO maybe handle pen size change directly within the display --> can be useful by the way
    def penSizechange(self):
        self.paint.brushSize = self.penSize.value()
        # self.update() # too slow if I update --> avoid
        # self.l1.setFont(QFont("Arial", size))

    def set_image(self, img):

        # probably I need store raw image to avoid issues --> TODO ???
        self.paint.set_image(img)# bug seems to be here

        img = self.paint.raw_image
        # print(type(img))

        # else I need also reset channels

        # if img is not None:
        #     if img.has_c():
        self._update_channels(img=img)
        # print('out')
        # channels = self.paint.get_nb_channels()
        # self._update_channels(channels)
        # make it update also the channels

    def set_mask(self, mask):
        self.paint.set_mask(mask)

    def _update_channels(self, img):
        selection = self.channels.currentIndex()
        self.channels.disconnect()
        self.channels.clear()
        comboData = ['merge']
        if img is not None:
            try:
                if img.has_c():
                    for i in range(img.get_dimension('c')):
                        comboData.append(str(i))
            except:
                # assume image has no channel and return None
                # or assume last stuff is channel if image has more than 2
                pass
        # logger.debug('channels found ' + str(comboData))
        self.channels.addItems(comboData)
        # index = self.channels.findData(selection)

        self.channels.currentIndexChanged.connect(self.channelChange)
        # print("data", index)
        if selection != -1 and selection < self.channels.count():
            self.channels.setCurrentIndex(selection)
            # self.channelChange(selection)
        else:
            self.channels.setCurrentIndex(0)
            # self.channelChange(0)


    def freeze(self, bool, level=1):

        self.tb1.setEnabled(True)
        self.tb2.setEnabled(True)
        self.tb3.setEnabled(True)

        if level == 1:
            self.tb2.setEnabled(not bool)
        elif level == 2:
            self.tb2.setEnabled(not bool)
            self.tb3.setEnabled(not bool)

        if bool:
            self.disable_shortcuts()
        else:
            self.enable_shortcuts()
        # remove draw mode maybe and freeze shortcuts

# pas mal faut juste voir ce que je peux faire
# peut etre permetre de mettre des custom panels

if __name__ == '__main__':
    import sys
    from qtpy.QtWidgets import QApplication


    # hand drawing panel
    if False:

        # TODO add a main method so it can be called directly
        # maybe just show a canvas and give it interesting props --> TODO --> really need fix that too!!!


        # should probably have his own scroll bar embedded somewhere

        app = QApplication(sys.argv)


        # quite easy todo in fact

        # no clue why it always works just once


        # maybe make a scrollable TA paint that will handle and behave like that
        # maybe handle also stacks at some point ---> see how
        # --> almost all is ok now
        class overriding_apply(Createpaintwidget):
            # all seems ok now and functions as in TA
            # just do the shift enter to get rid of small cells

            # avec ça ça marche 100% à la TA ... --> cool --> maybe make it a TA drawing pad
            def apply(self):
               self.apply_drawing(minimal_cell_size=0)

            def shift_apply(self):
                # MEGA TODO IMPLEMENT SIZE within this stuff!!!
                self.apply_drawing(minimal_cell_size=10)

            def ctrl_m_apply(self):
                self.manually_reseeded_wshed()

            def save(self):
                self.save_mask()

        w = scrollable_paint(
            custom_paint_panel=overriding_apply())  # ça marche --> permet de mettre des paint panels avec des proprietes particulieres --> assez facile en fait
        # w.set_image('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012.png')
        # w.set_mask('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012/handCorrection.png')
        # w.set_image('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/Optimized_projection_018.png')
        # w.set_mask('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/neo_mask.tif')

        w.set_image('/D/VirtualBox/book_chapter_image_segmentation_final_2021/new_figs/dataset_1/minimal_demo_sample_simplified_bckup_before_messing_with_two_seg_errors/1.tif')
        w.set_mask('/D/VirtualBox/book_chapter_image_segmentation_final_2021/new_figs/dataset_1/minimal_demo_sample_simplified_bckup_before_messing_with_two_seg_errors/1/handCorrection.tif')


        # ça marche mais tt finalizer
        # that seems to work

        # w.freeze(True)
        # w.freeze(False)
        # w.freeze(True)

        # all is so easy to do this way

        # ça marche --> tres facile a tester
        # --> how can I do a zoom
        # ask whether self scroll or not --> if scrollable

        # mask = w.get_mask()
        #
        # plt.imshow(mask)
        # plt.show()

        w.show()
        sys.exit(app.exec_())

    # vectorial panel for drawing ROIs --> that works --> keep this code cause very useful and many bug fix
    # TODO try convert any ROIs to something like a contour using the pavlidis algo --> put this in rps tools

    if True:
        # TODO add a main method so it can be called directly
        # maybe just show a canvas and give it interesting props --> TODO --> really need fix that too!!!

        # should probably have his own scroll bar embedded somewhere

        app = QApplication(sys.argv)


        # quite easy todo in fact

        # no clue why it always works just once

        # maybe make a scrollable TA paint that will handle and behave like that
        # maybe handle also stacks at some point ---> see how
        # --> almost all is ok now
        class overriding_apply(Createpaintwidget):
            # all seems ok now and functions as in TA
            # just do the shift enter to get rid of small cells
            def m_apply(self):
                self.vdp.active = not self.vdp.active
                self.update()

            # avec ça ça marche 100% à la TA ... --> cool --> maybe make it a TA drawing pad
            def apply(self):
                # self.apply_drawing(minimal_cell_size=0)
                pass

            def shift_apply(self):
                # MEGA TODO IMPLEMENT SIZE within this stuff!!!
                # self.apply_drawing(minimal_cell_size=10)
                pass

            def ctrl_m_apply(self):
                # print('seed')
                # self.manually_reseeded_wshed()
                pass

            def save(self):
                # self.save_mask()
                print('saving ROIs')
                print(self.vdp.shapes)
                # create an image the save size as the displayed one and open it
                if self.raw_image is None:
                    return
                if len(self.raw_image.shape)==2:
                    black_image = np.zeros_like(self.raw_image, dtype=np.uint64)
                else:
                    black_image = np.zeros(shape=(*self.raw_image[:-1],), dtype=np.uint64)
                # plt.imshow(black_image)
                # plt.show()
                # make it fill all the ROIs inside that
                im2d = Image2D(black_image)

                for iii,shape in enumerate(self.vdp.shapes):
                    # ideally need clone it and make it the enumerate color +1

                    shape.fill_color = iii+1
                    shape.color = None
                    # print(shape.fill_color)
                    # print(type(shape), shape.fill_color)
                    im2d.annotation.append(shape)



                # im2d.annotation.extend(self.vdp.shapes)

                # need clone the shape

                # I need remove antialiasing
                im = im2d.save(None,quality=QPainter.NonCosmeticDefaultPen)
                im = im2d.convert_qimage_to_numpy(im)

                # restore colors
                for shape in self.vdp.shapes:
                    shape.fill_color = None
                    shape.color = 0xFFFF00

                # faudrait tester en grandeur reelle pr voir si ça marche --> à faire !!!
                if True:
                    # TODO just save the raw file in one channel not in many --> check the  output by the way

                    # from skimage.measure import label
                    # lab = label(im, connectivity=2, background=0)
                    # im=lab

                    # I need remove antialiasing
                    Img(im.astype(np.float64), dimensions='hwc').save('/E/Sample_images/sample_image_neurons_laurence_had/data_test/test_raw.tif')

                    plt.imshow(im)
                    plt.show()

                # what I need to do here now is to save in the folder of the parent à la TA --> can save it as neurons.tif maybe




            # def suppr_pressed(self):
            #     # self.save_mask()
            #     # print('suppr')
            #     self.pa


        test = overriding_apply()
        test.vdp.shape_to_draw = Freehand2D

        # test.vdp.shape_to_draw = Rect2D # ça ça marche --> bug is in the other then...

        # there is a bug in the vectorial but no big deal probably

        w = scrollable_paint(
            custom_paint_panel=test)  # ça marche --> permet de mettre des paint panels avec des proprietes particulieres --> assez facile en fait
        # w.set_image('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012.png')
        # w.set_mask('/E/Sample_images/sample_images_PA/trash_test_mem/mini (copie)/focused_Series012/handCorrection.png')
        # w.set_image('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/Optimized_projection_018.png')
        # w.set_mask('/E/Sample_images/sample_images_PA/test_complete_wing_raphael/neo_mask.tif')

        test.vdp.active = True
        test.vdp.drawing_mode = True

        # QApplication.setOverrideCursor(Qt.WaitCursor)
        # do lengthy process
        # QApplication.restoreOverrideCursor()

        w.set_image(
            '/E/Sample_images/sample_image_neurons_laurence_had/data_test/image.png')
        # w.set_mask('/D/VirtualBox/book_chapter_image_segmentation_final_2021/new_figs/dataset_1/minimal_demo_sample_simplified_bckup_before_messing_with_two_seg_errors/1/handCorrection.tif')

        # w.setCursor(Qt.BlankCursor)

        # ça marche mais tt finalizer
        # that seems to work

        # w.freeze(True)
        # w.freeze(False)
        # w.freeze(True)

        # we force the cursor to be visible --> set it by default
        # cursor = Qt.ArrowCursor
        # w.paint.setCursor(cursor)

        # all is so easy to do this way

        # ça marche --> tres facile a tester
        # --> how can I do a zoom
        # ask whether self scroll or not --> if scrollable

        # mask = w.get_mask()
        #
        # plt.imshow(mask)
        # plt.show()

        w.show()
        # w.setCursor(Qt.BlankCursor)
        sys.exit(app.exec_())



import os
from epyseg.settings.global_settings import set_UI # set the UI to be used py qtpy
set_UI()
from qtpy import QtCore
from epyseg.draw.shapes.polygon2d import Polygon2D
from epyseg.draw.shapes.line2d import Line2D
from epyseg.draw.shapes.rect2d import Rect2D
from epyseg.draw.shapes.square2d import Square2D
from epyseg.draw.shapes.ellipse2d import Ellipse2D
from epyseg.draw.shapes.circle2d import Circle2D
from epyseg.draw.shapes.freehand2d import Freehand2D
from epyseg.draw.shapes.point2d import Point2D
from epyseg.draw.shapes.polyline2d import PolyLine2D
from epyseg.draw.shapes.image2d import Image2D
from qtpy.QtCore import QPointF, QRectF
from epyseg.tools.logger import TA_logger # logging
logger = TA_logger()

class VectorialDrawPane:

    def __init__(self, active=False, demo=False, scale=1.0, drawing_mode=False):
        self.shapes = []
        self.currently_drawn_shape = None
        self.shape_to_draw = None
        self.selected_shape = []
        self.active = active
        self.scale = scale
        self.drawing_mode = drawing_mode
        if demo:
            self.shapes.append(Polygon2D(0, 0, 10, 0, 10, 20, 0, 20, 0, 0, color=0x00FF00))
            self.shapes.append(
                Polygon2D(100, 100, 110, 100, 110, 120, 10, 120, 100, 100, color=0x0000FF, fill_color=0x00FFFF,
                          stroke=2))
            self.shapes.append(Line2D(0, 0, 110, 100, color=0xFF0000, stroke=3))
            self.shapes.append(Rect2D(200, 150, 250, 100, stroke=10))
            self.shapes.append(Square2D(300, 260, 250, stroke=3))
            self.shapes.append(Ellipse2D(0, 50, 600, 200, stroke=3))
            self.shapes.append(Circle2D(150, 300, 30, color=0xFF0000))
            self.shapes.append(PolyLine2D(10, 10, 20, 10, 20, 30, 40, 30, color=0xFF0000, stroke=2))
            self.shapes.append(PolyLine2D(10, 10, 20, 10, 20, 30, 40, 30, color=0xFF0000, stroke=2))
            self.shapes.append(Point2D(128, 128, color=0xFF0000, stroke=6))
            self.shapes.append(Point2D(128, 128, color=0x00FF00, stroke=1))
            self.shapes.append(Point2D(10, 10, color=0x000000, stroke=6))
            img0 = Image2D('./../data/counter/00.png')
            img1 = Image2D('./../data/counter/01.png')
            img2 = Image2D('./../data/counter/02.png')
            img3 = Image2D('./../data/counter/03.png')
            img4 = Image2D('./../data/counter/04.png')
            img5 = Image2D('./../data/counter/05.png')
            img6 = Image2D('./../data/counter/06.png')
            img7 = Image2D('./../data/counter/07.png')
            img8 = Image2D('./../data/counter/08.png')
            img9 = Image2D('./../data/counter/09.png')
            img10 = Image2D('./../data/counter/10.png')

            row = img1 + img2 + img10

            self.shapes.append(row)

            row2 = img4 + img5
            fig = row / row2
            # fig = Column(row, row2)
            #self.shapes.append(fig)
            self.drawing_mode = True
            # self.shape_to_draw = Line2D
            # self.shape_to_draw = Rect2D
            # self.shape_to_draw = Square2D
            # self.shape_to_draw = Ellipse2D
            # self.shape_to_draw = Circle2D
            # self.shape_to_draw = Point2D  # ok maybe small centering issue
            # self.shape_to_draw = Freehand2D
            # self.shape_to_draw = PolyLine2D
            # self.shape_to_draw = Polygon2D
            import random
            drawing_methods = [Line2D, Rect2D, Square2D, Ellipse2D, Circle2D, Point2D, Freehand2D, PolyLine2D, Polygon2D]
            self.shape_to_draw = random.choice(drawing_methods)

            # TODO freehand drawing
            # TODO broken line --> need double click for end

    def paintEvent(self, *args):
        painter = args[0]
        visibleRect = None
        if len(args) >= 2:
              visibleRect = args[1]

        painter.save()
        if self.scale != 1.0:
            painter.scale(self.scale, self.scale)

        for shape in self.shapes:
            # only draw shapes if they are visible --> requires a visiblerect to be passed
            if visibleRect is not None:
                # only draws if in visible rect
                if shape.boundingRect().intersects(QRectF(visibleRect)):
                    shape.draw(painter)
            else:
                shape.draw(painter)

        if self.currently_drawn_shape is not None:
            if self.currently_drawn_shape.isSet:
                self.currently_drawn_shape.draw(painter)

        sel = self.create_master_rect()
        if sel is not None:
            painter.drawRect(sel)
        painter.restore()
        # painter.end() # probably a good idea ????

    def group_contains(self, x, y):
        # checks if master rect for group contains click
        # get bounds and create union and compare
        master_rect = self.create_master_rect()
        if master_rect is None:
            return False
        return master_rect.contains(QPointF(x, y))

    def create_master_rect(self):
        master_rect = None
        if self.selected_shape:
            for shape in self.selected_shape:
                if master_rect is None:
                    master_rect = shape.boundingRect()
                else:
                    master_rect = master_rect.united(shape.boundingRect())
        return master_rect

    def removeCurShape(self):
        if self.selected_shape:
            self.shapes = [e for e in self.shapes if e not in self.selected_shape]
            self.selected_shape = []

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.drawing = True
            self.lastPoint = event.pos() / self.scale
            self.firstPoint = event.pos() / self.scale

            shapeFound = False
            if self.currently_drawn_shape is None:
                for shape in reversed(self.shapes):
                    if shape.contains(self.lastPoint) and not shape in self.selected_shape:
                        logger.debug('you clicked shape:' + str(shape))
                        if event.modifiers() == QtCore.Qt.ControlModifier:
                            if shape not in self.selected_shape:  # avoid doublons
                                self.selected_shape.append(shape)  # add shape to group
                                logger.debug('adding shape to group')
                                shapeFound = True
                        else:
                            if not self.group_contains(self.lastPoint.x(), self.lastPoint.y()):
                                self.selected_shape = [shape]
                                logger.debug('only one element is selected')
                                shapeFound = True
                        return

                if not shapeFound and event.modifiers() == QtCore.Qt.ControlModifier:
                    for shape in reversed(self.shapes):
                        if shape.contains(self.lastPoint):
                            if shape in self.selected_shape:  # avoid doublons
                                logger.debug('you clicked again shape:' + str(shape))
                                self.selected_shape.remove(shape)  # add shape to group
                                logger.debug('removing a shape from group')
                                shapeFound = True
                # no shape found --> reset sel
                if not shapeFound and not self.group_contains(self.lastPoint.x(), self.lastPoint.y()):
                    logger.debug('resetting sel')
                    self.selected_shape = []

            # check if a shape is selected and only move that
            if self.drawing_mode and not self.selected_shape and self.currently_drawn_shape is None:
                # do not reset shape if not done drawing...
                if self.shape_to_draw is not None:
                    self.currently_drawn_shape = self.shape_to_draw()
                else:
                    self.currently_drawn_shape = None
            if self.drawing_mode and not self.selected_shape:
                if self.currently_drawn_shape is not None:
                    self.currently_drawn_shape.set_P1(QPointF(self.lastPoint.x(), self.lastPoint.y()))

    def mouseMoveEvent(self, event):
        if event.buttons() and QtCore.Qt.LeftButton:
            if self.selected_shape and self.currently_drawn_shape is None:
                logger.debug('moving' + str(self.selected_shape))
                for shape in self.selected_shape:
                    shape.translate(event.pos() / self.scale - self.lastPoint)

        if self.currently_drawn_shape is not None:
            self.currently_drawn_shape.add(self.firstPoint, self.lastPoint)

        self.lastPoint = event.pos() / self.scale

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.drawing = False
            if self.drawing_mode and self.currently_drawn_shape is not None:
                self.currently_drawn_shape.add(self.firstPoint, self.lastPoint)
                if isinstance(self.currently_drawn_shape, Freehand2D):
                    # this closes the freehand shape
                    self.currently_drawn_shape.add(self.lastPoint, self.firstPoint)
                # should not erase the shape if it's a polyline or a polygon by the way
                if isinstance(self.currently_drawn_shape, Freehand2D) or (not isinstance(self.currently_drawn_shape, PolyLine2D) and not isinstance(self.currently_drawn_shape, Polygon2D)):
                    self.shapes.append(self.currently_drawn_shape)
                    self.currently_drawn_shape = None

    def mouseDoubleClickEvent(self, event):
        if isinstance(self.currently_drawn_shape, PolyLine2D) or isinstance(self.currently_drawn_shape, Polygon2D):
            self.shapes.append(self.currently_drawn_shape)
            self.currently_drawn_shape = None

if __name__ == '__main__':
    VectorialDrawPane()



import math
from random import gauss

import matplotlib.pyplot as plt
import numpy as np
import qtpy
from numpy.ma.testutils import assert_array_equal

from epyseg.img import Img,  pop
from epyseg.settings.global_settings import print_default_qtpy_UI_really_in_use
from epyseg.utils.loadlist import loadlist
import random


# do that for all GUIs --> TODO

if __name__ == '__main__':
    if True:
        # incroyable ça marche ici mais pas en unittest ... --> comprend rien

        # KEEP ALL ça marche
        # tst_pyqt5()
        # tst_pyside2()
        # tst_pyside6()
        # there are weird bugs but globally ok


        import sys
        sys.exit(0)

    if True:
        img = Img('/E/Sample_images/sample_images_FIJI/150707_WTstack.lsm')
        # auto_scale(img)
        pop(img)

        img = Img('/E/Sample_images/sample_images_FIJI/AuPbSn40.jpg')
        pop(img)
        # ok but then where is the bug???

        import sys
        sys.exit(0)
